from google.protobuf import struct_pb2 as _struct_pb2
from google.protobuf import wrappers_pb2 as _wrappers_pb2
from qm.grpc.qm.grpc.v2 import qm_api_pb2 as _qm_api_pb2
from qm.grpc.qm.grpc.v2 import common_types_pb2 as _common_types_pb2
from qm.grpc.qm.pb import inc_qm_api_pb2 as _inc_qm_api_pb2
from qm.grpc.qm.pb import job_manager_pb2 as _job_manager_pb2
from qm.grpc.qm.pb import job_results_pb2 as _job_results_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class SetMatrixCorrectionResponse(_message.Message):
    __slots__ = ("success", "error")
    class SetMatrixCorrectionResponseError(_message.Message):
        __slots__ = ("details",)
        DETAILS_FIELD_NUMBER: _ClassVar[int]
        details: str
        def __init__(self, details: _Optional[str] = ...) -> None: ...
    class SetMatrixCorrectionResponseSuccess(_message.Message):
        __slots__ = ()
        def __init__(self) -> None: ...
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    success: SetMatrixCorrectionResponse.SetMatrixCorrectionResponseSuccess
    error: SetMatrixCorrectionResponse.SetMatrixCorrectionResponseError
    def __init__(self, success: _Optional[_Union[SetMatrixCorrectionResponse.SetMatrixCorrectionResponseSuccess, _Mapping]] = ..., error: _Optional[_Union[SetMatrixCorrectionResponse.SetMatrixCorrectionResponseError, _Mapping]] = ...) -> None: ...

class SetIntermediateFrequencyResponse(_message.Message):
    __slots__ = ("success", "error")
    class SetIntermediateFrequencyResponseError(_message.Message):
        __slots__ = ("details",)
        DETAILS_FIELD_NUMBER: _ClassVar[int]
        details: str
        def __init__(self, details: _Optional[str] = ...) -> None: ...
    class SetIntermediateFrequencyResponseSuccess(_message.Message):
        __slots__ = ()
        def __init__(self) -> None: ...
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    success: SetIntermediateFrequencyResponse.SetIntermediateFrequencyResponseSuccess
    error: SetIntermediateFrequencyResponse.SetIntermediateFrequencyResponseError
    def __init__(self, success: _Optional[_Union[SetIntermediateFrequencyResponse.SetIntermediateFrequencyResponseSuccess, _Mapping]] = ..., error: _Optional[_Union[SetIntermediateFrequencyResponse.SetIntermediateFrequencyResponseError, _Mapping]] = ...) -> None: ...

class SetDigitalDelayResponse(_message.Message):
    __slots__ = ("success", "error")
    class SetDigitalDelayResponseError(_message.Message):
        __slots__ = ("details",)
        DETAILS_FIELD_NUMBER: _ClassVar[int]
        details: str
        def __init__(self, details: _Optional[str] = ...) -> None: ...
    class SetDigitalDelayResponseSuccess(_message.Message):
        __slots__ = ()
        def __init__(self) -> None: ...
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    success: SetDigitalDelayResponse.SetDigitalDelayResponseSuccess
    error: SetDigitalDelayResponse.SetDigitalDelayResponseError
    def __init__(self, success: _Optional[_Union[SetDigitalDelayResponse.SetDigitalDelayResponseSuccess, _Mapping]] = ..., error: _Optional[_Union[SetDigitalDelayResponse.SetDigitalDelayResponseError, _Mapping]] = ...) -> None: ...

class SetDigitalBufferResponse(_message.Message):
    __slots__ = ("success", "error")
    class SetDigitalBufferResponseError(_message.Message):
        __slots__ = ("details",)
        DETAILS_FIELD_NUMBER: _ClassVar[int]
        details: str
        def __init__(self, details: _Optional[str] = ...) -> None: ...
    class SetDigitalBufferResponseSuccess(_message.Message):
        __slots__ = ()
        def __init__(self) -> None: ...
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    success: SetDigitalBufferResponse.SetDigitalBufferResponseSuccess
    error: SetDigitalBufferResponse.SetDigitalBufferResponseError
    def __init__(self, success: _Optional[_Union[SetDigitalBufferResponse.SetDigitalBufferResponseSuccess, _Mapping]] = ..., error: _Optional[_Union[SetDigitalBufferResponse.SetDigitalBufferResponseError, _Mapping]] = ...) -> None: ...

class SetInputDcOffsetResponse(_message.Message):
    __slots__ = ("success", "error")
    class SetInputDcOffsetResponseError(_message.Message):
        __slots__ = ("details",)
        DETAILS_FIELD_NUMBER: _ClassVar[int]
        details: str
        def __init__(self, details: _Optional[str] = ...) -> None: ...
    class SetInputDcOffsetResponseSuccess(_message.Message):
        __slots__ = ()
        def __init__(self) -> None: ...
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    success: SetInputDcOffsetResponse.SetInputDcOffsetResponseSuccess
    error: SetInputDcOffsetResponse.SetInputDcOffsetResponseError
    def __init__(self, success: _Optional[_Union[SetInputDcOffsetResponse.SetInputDcOffsetResponseSuccess, _Mapping]] = ..., error: _Optional[_Union[SetInputDcOffsetResponse.SetInputDcOffsetResponseError, _Mapping]] = ...) -> None: ...

class SetOutputDcOffsetResponse(_message.Message):
    __slots__ = ("success", "error")
    class SetOutputDcOffsetResponseError(_message.Message):
        __slots__ = ("details",)
        DETAILS_FIELD_NUMBER: _ClassVar[int]
        details: str
        def __init__(self, details: _Optional[str] = ...) -> None: ...
    class SetOutputDcOffsetResponseSuccess(_message.Message):
        __slots__ = ()
        def __init__(self) -> None: ...
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    success: SetOutputDcOffsetResponse.SetOutputDcOffsetResponseSuccess
    error: SetOutputDcOffsetResponse.SetOutputDcOffsetResponseError
    def __init__(self, success: _Optional[_Union[SetOutputDcOffsetResponse.SetOutputDcOffsetResponseSuccess, _Mapping]] = ..., error: _Optional[_Union[SetOutputDcOffsetResponse.SetOutputDcOffsetResponseError, _Mapping]] = ...) -> None: ...

class SetIoValuesResponse(_message.Message):
    __slots__ = ("success", "error")
    class SetIoValuesResponseError(_message.Message):
        __slots__ = ("details",)
        DETAILS_FIELD_NUMBER: _ClassVar[int]
        details: str
        def __init__(self, details: _Optional[str] = ...) -> None: ...
    class SetIoValuesResponseSuccess(_message.Message):
        __slots__ = ()
        def __init__(self) -> None: ...
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    success: SetIoValuesResponse.SetIoValuesResponseSuccess
    error: SetIoValuesResponse.SetIoValuesResponseError
    def __init__(self, success: _Optional[_Union[SetIoValuesResponse.SetIoValuesResponseSuccess, _Mapping]] = ..., error: _Optional[_Union[SetIoValuesResponse.SetIoValuesResponseError, _Mapping]] = ...) -> None: ...

class SetIoValuesRequest(_message.Message):
    __slots__ = ("job_id", "io1", "io2")
    class IOValueSetData(_message.Message):
        __slots__ = ("int_value", "double_value", "boolean_value")
        INT_VALUE_FIELD_NUMBER: _ClassVar[int]
        DOUBLE_VALUE_FIELD_NUMBER: _ClassVar[int]
        BOOLEAN_VALUE_FIELD_NUMBER: _ClassVar[int]
        int_value: int
        double_value: float
        boolean_value: bool
        def __init__(self, int_value: _Optional[int] = ..., double_value: _Optional[float] = ..., boolean_value: bool = ...) -> None: ...
    JOB_ID_FIELD_NUMBER: _ClassVar[int]
    IO1_FIELD_NUMBER: _ClassVar[int]
    IO2_FIELD_NUMBER: _ClassVar[int]
    job_id: str
    io1: SetIoValuesRequest.IOValueSetData
    io2: SetIoValuesRequest.IOValueSetData
    def __init__(self, job_id: _Optional[str] = ..., io1: _Optional[_Union[SetIoValuesRequest.IOValueSetData, _Mapping]] = ..., io2: _Optional[_Union[SetIoValuesRequest.IOValueSetData, _Mapping]] = ...) -> None: ...

class GetIoValuesRequest(_message.Message):
    __slots__ = ("job_id",)
    JOB_ID_FIELD_NUMBER: _ClassVar[int]
    job_id: str
    def __init__(self, job_id: _Optional[str] = ...) -> None: ...

class IOValuesData(_message.Message):
    __slots__ = ("int_value", "double_value", "boolean_value")
    INT_VALUE_FIELD_NUMBER: _ClassVar[int]
    DOUBLE_VALUE_FIELD_NUMBER: _ClassVar[int]
    BOOLEAN_VALUE_FIELD_NUMBER: _ClassVar[int]
    int_value: int
    double_value: float
    boolean_value: bool
    def __init__(self, int_value: _Optional[int] = ..., double_value: _Optional[float] = ..., boolean_value: bool = ...) -> None: ...

class GetIoValuesResponse(_message.Message):
    __slots__ = ("success", "error")
    class GetIoValuesResponseSuccess(_message.Message):
        __slots__ = ("io1", "io2")
        class IOValuesData(_message.Message):
            __slots__ = ("int_value", "double_value", "boolean_value")
            INT_VALUE_FIELD_NUMBER: _ClassVar[int]
            DOUBLE_VALUE_FIELD_NUMBER: _ClassVar[int]
            BOOLEAN_VALUE_FIELD_NUMBER: _ClassVar[int]
            int_value: int
            double_value: float
            boolean_value: bool
            def __init__(self, int_value: _Optional[int] = ..., double_value: _Optional[float] = ..., boolean_value: bool = ...) -> None: ...
        IO1_FIELD_NUMBER: _ClassVar[int]
        IO2_FIELD_NUMBER: _ClassVar[int]
        io1: GetIoValuesResponse.GetIoValuesResponseSuccess.IOValuesData
        io2: GetIoValuesResponse.GetIoValuesResponseSuccess.IOValuesData
        def __init__(self, io1: _Optional[_Union[GetIoValuesResponse.GetIoValuesResponseSuccess.IOValuesData, _Mapping]] = ..., io2: _Optional[_Union[GetIoValuesResponse.GetIoValuesResponseSuccess.IOValuesData, _Mapping]] = ...) -> None: ...
    class GetIoValuesResponseError(_message.Message):
        __slots__ = ("details",)
        DETAILS_FIELD_NUMBER: _ClassVar[int]
        details: str
        def __init__(self, details: _Optional[str] = ...) -> None: ...
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    success: GetIoValuesResponse.GetIoValuesResponseSuccess
    error: GetIoValuesResponse.GetIoValuesResponseError
    def __init__(self, success: _Optional[_Union[GetIoValuesResponse.GetIoValuesResponseSuccess, _Mapping]] = ..., error: _Optional[_Union[GetIoValuesResponse.GetIoValuesResponseError, _Mapping]] = ...) -> None: ...

class JobServiceIsPausedRequest(_message.Message):
    __slots__ = ("job_id",)
    JOB_ID_FIELD_NUMBER: _ClassVar[int]
    job_id: str
    def __init__(self, job_id: _Optional[str] = ...) -> None: ...

class IsPausedResponse(_message.Message):
    __slots__ = ("error", "success")
    class IsPausedResponseError(_message.Message):
        __slots__ = ("details",)
        DETAILS_FIELD_NUMBER: _ClassVar[int]
        details: str
        def __init__(self, details: _Optional[str] = ...) -> None: ...
    class IsPausedResponseSuccess(_message.Message):
        __slots__ = ("is_paused",)
        IS_PAUSED_FIELD_NUMBER: _ClassVar[int]
        is_paused: bool
        def __init__(self, is_paused: bool = ...) -> None: ...
    ERROR_FIELD_NUMBER: _ClassVar[int]
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    error: IsPausedResponse.IsPausedResponseError
    success: IsPausedResponse.IsPausedResponseSuccess
    def __init__(self, error: _Optional[_Union[IsPausedResponse.IsPausedResponseError, _Mapping]] = ..., success: _Optional[_Union[IsPausedResponse.IsPausedResponseSuccess, _Mapping]] = ...) -> None: ...

class JobServiceGetConfigRequest(_message.Message):
    __slots__ = ("job_id",)
    JOB_ID_FIELD_NUMBER: _ClassVar[int]
    job_id: str
    def __init__(self, job_id: _Optional[str] = ...) -> None: ...

class JobServiceGetJobStatusRequest(_message.Message):
    __slots__ = ("job_id",)
    JOB_ID_FIELD_NUMBER: _ClassVar[int]
    job_id: str
    def __init__(self, job_id: _Optional[str] = ...) -> None: ...

class ResumeRequest(_message.Message):
    __slots__ = ("job_id",)
    JOB_ID_FIELD_NUMBER: _ClassVar[int]
    job_id: str
    def __init__(self, job_id: _Optional[str] = ...) -> None: ...

class ResumeResponse(_message.Message):
    __slots__ = ("success", "error")
    class ResumeResponseError(_message.Message):
        __slots__ = ("details",)
        DETAILS_FIELD_NUMBER: _ClassVar[int]
        details: str
        def __init__(self, details: _Optional[str] = ...) -> None: ...
    class ResumeResponseSuccess(_message.Message):
        __slots__ = ()
        def __init__(self) -> None: ...
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    success: ResumeResponse.ResumeResponseSuccess
    error: ResumeResponse.ResumeResponseError
    def __init__(self, success: _Optional[_Union[ResumeResponse.ResumeResponseSuccess, _Mapping]] = ..., error: _Optional[_Union[ResumeResponse.ResumeResponseError, _Mapping]] = ...) -> None: ...

class SetMatrixCorrectionRequest(_message.Message):
    __slots__ = ("job_id", "qe", "correction")
    JOB_ID_FIELD_NUMBER: _ClassVar[int]
    QE_FIELD_NUMBER: _ClassVar[int]
    CORRECTION_FIELD_NUMBER: _ClassVar[int]
    job_id: str
    qe: str
    correction: _inc_qm_api_pb2.Matrix
    def __init__(self, job_id: _Optional[str] = ..., qe: _Optional[str] = ..., correction: _Optional[_Union[_inc_qm_api_pb2.Matrix, _Mapping]] = ...) -> None: ...

class SetIntermediateFrequencyRequest(_message.Message):
    __slots__ = ("job_id", "qe", "frequency")
    JOB_ID_FIELD_NUMBER: _ClassVar[int]
    QE_FIELD_NUMBER: _ClassVar[int]
    FREQUENCY_FIELD_NUMBER: _ClassVar[int]
    job_id: str
    qe: str
    frequency: float
    def __init__(self, job_id: _Optional[str] = ..., qe: _Optional[str] = ..., frequency: _Optional[float] = ...) -> None: ...

class SetDigitalDelayRequest(_message.Message):
    __slots__ = ("job_id", "qe", "port", "delay")
    JOB_ID_FIELD_NUMBER: _ClassVar[int]
    QE_FIELD_NUMBER: _ClassVar[int]
    PORT_FIELD_NUMBER: _ClassVar[int]
    DELAY_FIELD_NUMBER: _ClassVar[int]
    job_id: str
    qe: str
    port: str
    delay: int
    def __init__(self, job_id: _Optional[str] = ..., qe: _Optional[str] = ..., port: _Optional[str] = ..., delay: _Optional[int] = ...) -> None: ...

class SetDigitalBufferRequest(_message.Message):
    __slots__ = ("job_id", "qe", "port", "buffer")
    JOB_ID_FIELD_NUMBER: _ClassVar[int]
    QE_FIELD_NUMBER: _ClassVar[int]
    PORT_FIELD_NUMBER: _ClassVar[int]
    BUFFER_FIELD_NUMBER: _ClassVar[int]
    job_id: str
    qe: str
    port: str
    buffer: int
    def __init__(self, job_id: _Optional[str] = ..., qe: _Optional[str] = ..., port: _Optional[str] = ..., buffer: _Optional[int] = ...) -> None: ...

class SetInputDcOffsetRequest(_message.Message):
    __slots__ = ("job_id", "qe", "port", "offset")
    JOB_ID_FIELD_NUMBER: _ClassVar[int]
    QE_FIELD_NUMBER: _ClassVar[int]
    PORT_FIELD_NUMBER: _ClassVar[int]
    OFFSET_FIELD_NUMBER: _ClassVar[int]
    job_id: str
    qe: str
    port: str
    offset: float
    def __init__(self, job_id: _Optional[str] = ..., qe: _Optional[str] = ..., port: _Optional[str] = ..., offset: _Optional[float] = ...) -> None: ...

class SetOutputDcOffsetRequest(_message.Message):
    __slots__ = ("job_id", "qe", "mix_inputs", "single_input")
    JOB_ID_FIELD_NUMBER: _ClassVar[int]
    QE_FIELD_NUMBER: _ClassVar[int]
    MIX_INPUTS_FIELD_NUMBER: _ClassVar[int]
    SINGLE_INPUT_FIELD_NUMBER: _ClassVar[int]
    job_id: str
    qe: str
    mix_inputs: MixInputsDcOffset
    single_input: SingleInputDcOffset
    def __init__(self, job_id: _Optional[str] = ..., qe: _Optional[str] = ..., mix_inputs: _Optional[_Union[MixInputsDcOffset, _Mapping]] = ..., single_input: _Optional[_Union[SingleInputDcOffset, _Mapping]] = ...) -> None: ...

class GetMatrixCorrectionRequest(_message.Message):
    __slots__ = ("job_id", "qe")
    JOB_ID_FIELD_NUMBER: _ClassVar[int]
    QE_FIELD_NUMBER: _ClassVar[int]
    job_id: str
    qe: str
    def __init__(self, job_id: _Optional[str] = ..., qe: _Optional[str] = ...) -> None: ...

class GetMatrixCorrectionError(_message.Message):
    __slots__ = ("details",)
    DETAILS_FIELD_NUMBER: _ClassVar[int]
    details: str
    def __init__(self, details: _Optional[str] = ...) -> None: ...

class GetMatrixCorrectionResponse(_message.Message):
    __slots__ = ("error", "success")
    class GetMatrixCorrectionResponseSuccess(_message.Message):
        __slots__ = ("correction",)
        CORRECTION_FIELD_NUMBER: _ClassVar[int]
        correction: _inc_qm_api_pb2.Matrix
        def __init__(self, correction: _Optional[_Union[_inc_qm_api_pb2.Matrix, _Mapping]] = ...) -> None: ...
    class GetMatrixCorrectionResponseError(_message.Message):
        __slots__ = ("error",)
        ERROR_FIELD_NUMBER: _ClassVar[int]
        error: GetMatrixCorrectionError
        def __init__(self, error: _Optional[_Union[GetMatrixCorrectionError, _Mapping]] = ...) -> None: ...
    ERROR_FIELD_NUMBER: _ClassVar[int]
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    error: GetMatrixCorrectionResponse.GetMatrixCorrectionResponseError
    success: GetMatrixCorrectionResponse.GetMatrixCorrectionResponseSuccess
    def __init__(self, error: _Optional[_Union[GetMatrixCorrectionResponse.GetMatrixCorrectionResponseError, _Mapping]] = ..., success: _Optional[_Union[GetMatrixCorrectionResponse.GetMatrixCorrectionResponseSuccess, _Mapping]] = ...) -> None: ...

class GetIntermediateFrequencyRequest(_message.Message):
    __slots__ = ("job_id", "qe")
    JOB_ID_FIELD_NUMBER: _ClassVar[int]
    QE_FIELD_NUMBER: _ClassVar[int]
    job_id: str
    qe: str
    def __init__(self, job_id: _Optional[str] = ..., qe: _Optional[str] = ...) -> None: ...

class GetIntermediateFrequencyResponse(_message.Message):
    __slots__ = ("error", "success")
    class GetIntermediateFrequencyResponseError(_message.Message):
        __slots__ = ("details",)
        DETAILS_FIELD_NUMBER: _ClassVar[int]
        details: str
        def __init__(self, details: _Optional[str] = ...) -> None: ...
    class GetIntermediateFrequencyResponseSuccess(_message.Message):
        __slots__ = ("frequency",)
        FREQUENCY_FIELD_NUMBER: _ClassVar[int]
        frequency: float
        def __init__(self, frequency: _Optional[float] = ...) -> None: ...
    ERROR_FIELD_NUMBER: _ClassVar[int]
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    error: GetIntermediateFrequencyResponse.GetIntermediateFrequencyResponseError
    success: GetIntermediateFrequencyResponse.GetIntermediateFrequencyResponseSuccess
    def __init__(self, error: _Optional[_Union[GetIntermediateFrequencyResponse.GetIntermediateFrequencyResponseError, _Mapping]] = ..., success: _Optional[_Union[GetIntermediateFrequencyResponse.GetIntermediateFrequencyResponseSuccess, _Mapping]] = ...) -> None: ...

class GetDigitalDelayRequest(_message.Message):
    __slots__ = ("job_id", "qe", "port")
    JOB_ID_FIELD_NUMBER: _ClassVar[int]
    QE_FIELD_NUMBER: _ClassVar[int]
    PORT_FIELD_NUMBER: _ClassVar[int]
    job_id: str
    qe: str
    port: str
    def __init__(self, job_id: _Optional[str] = ..., qe: _Optional[str] = ..., port: _Optional[str] = ...) -> None: ...

class GetDigitalDelayResponse(_message.Message):
    __slots__ = ("success", "error")
    class GetDigitalDelayResponseError(_message.Message):
        __slots__ = ("details",)
        DETAILS_FIELD_NUMBER: _ClassVar[int]
        details: str
        def __init__(self, details: _Optional[str] = ...) -> None: ...
    class GetDigitalDelayResponseSuccess(_message.Message):
        __slots__ = ("delay",)
        DELAY_FIELD_NUMBER: _ClassVar[int]
        delay: int
        def __init__(self, delay: _Optional[int] = ...) -> None: ...
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    success: GetDigitalDelayResponse.GetDigitalDelayResponseSuccess
    error: GetDigitalDelayResponse.GetDigitalDelayResponseError
    def __init__(self, success: _Optional[_Union[GetDigitalDelayResponse.GetDigitalDelayResponseSuccess, _Mapping]] = ..., error: _Optional[_Union[GetDigitalDelayResponse.GetDigitalDelayResponseError, _Mapping]] = ...) -> None: ...

class GetDigitalBufferRequest(_message.Message):
    __slots__ = ("job_id", "qe", "port")
    JOB_ID_FIELD_NUMBER: _ClassVar[int]
    QE_FIELD_NUMBER: _ClassVar[int]
    PORT_FIELD_NUMBER: _ClassVar[int]
    job_id: str
    qe: str
    port: str
    def __init__(self, job_id: _Optional[str] = ..., qe: _Optional[str] = ..., port: _Optional[str] = ...) -> None: ...

class GetDigitalBufferResponse(_message.Message):
    __slots__ = ("success", "error")
    class GetDigitalBufferResponseError(_message.Message):
        __slots__ = ("details",)
        DETAILS_FIELD_NUMBER: _ClassVar[int]
        details: str
        def __init__(self, details: _Optional[str] = ...) -> None: ...
    class GetDigitalBufferResponseSuccess(_message.Message):
        __slots__ = ("buffer",)
        BUFFER_FIELD_NUMBER: _ClassVar[int]
        buffer: int
        def __init__(self, buffer: _Optional[int] = ...) -> None: ...
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    success: GetDigitalBufferResponse.GetDigitalBufferResponseSuccess
    error: GetDigitalBufferResponse.GetDigitalBufferResponseError
    def __init__(self, success: _Optional[_Union[GetDigitalBufferResponse.GetDigitalBufferResponseSuccess, _Mapping]] = ..., error: _Optional[_Union[GetDigitalBufferResponse.GetDigitalBufferResponseError, _Mapping]] = ...) -> None: ...

class GetInputDcOffsetRequest(_message.Message):
    __slots__ = ("job_id", "qe", "port")
    JOB_ID_FIELD_NUMBER: _ClassVar[int]
    QE_FIELD_NUMBER: _ClassVar[int]
    PORT_FIELD_NUMBER: _ClassVar[int]
    job_id: str
    qe: str
    port: str
    def __init__(self, job_id: _Optional[str] = ..., qe: _Optional[str] = ..., port: _Optional[str] = ...) -> None: ...

class GetInputDcOffsetError(_message.Message):
    __slots__ = ("details",)
    DETAILS_FIELD_NUMBER: _ClassVar[int]
    details: str
    def __init__(self, details: _Optional[str] = ...) -> None: ...

class GetInputDcOffsetSuccess(_message.Message):
    __slots__ = ("offset",)
    OFFSET_FIELD_NUMBER: _ClassVar[int]
    offset: float
    def __init__(self, offset: _Optional[float] = ...) -> None: ...

class GetInputDcOffsetResponse(_message.Message):
    __slots__ = ("success", "error")
    class GetInputDcOffsetResponseError(_message.Message):
        __slots__ = ("details",)
        DETAILS_FIELD_NUMBER: _ClassVar[int]
        details: str
        def __init__(self, details: _Optional[str] = ...) -> None: ...
    class GetInputDcOffsetResponseSuccess(_message.Message):
        __slots__ = ("offset",)
        OFFSET_FIELD_NUMBER: _ClassVar[int]
        offset: float
        def __init__(self, offset: _Optional[float] = ...) -> None: ...
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    success: GetInputDcOffsetResponse.GetInputDcOffsetResponseSuccess
    error: GetInputDcOffsetResponse.GetInputDcOffsetResponseError
    def __init__(self, success: _Optional[_Union[GetInputDcOffsetResponse.GetInputDcOffsetResponseSuccess, _Mapping]] = ..., error: _Optional[_Union[GetInputDcOffsetResponse.GetInputDcOffsetResponseError, _Mapping]] = ...) -> None: ...

class GetOutputDcOffsetRequest(_message.Message):
    __slots__ = ("job_id", "qe")
    JOB_ID_FIELD_NUMBER: _ClassVar[int]
    QE_FIELD_NUMBER: _ClassVar[int]
    job_id: str
    qe: str
    def __init__(self, job_id: _Optional[str] = ..., qe: _Optional[str] = ...) -> None: ...

class GetOutputDcOffsetResponse(_message.Message):
    __slots__ = ("success", "error")
    class GetOutputDcOffsetResponseError(_message.Message):
        __slots__ = ("details",)
        DETAILS_FIELD_NUMBER: _ClassVar[int]
        details: str
        def __init__(self, details: _Optional[str] = ...) -> None: ...
    class GetOutputDcOffsetResponseSuccess(_message.Message):
        __slots__ = ("mix_inputs", "single_input")
        MIX_INPUTS_FIELD_NUMBER: _ClassVar[int]
        SINGLE_INPUT_FIELD_NUMBER: _ClassVar[int]
        mix_inputs: MixInputsDcOffset
        single_input: SingleInputDcOffset
        def __init__(self, mix_inputs: _Optional[_Union[MixInputsDcOffset, _Mapping]] = ..., single_input: _Optional[_Union[SingleInputDcOffset, _Mapping]] = ...) -> None: ...
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    success: GetOutputDcOffsetResponse.GetOutputDcOffsetResponseSuccess
    error: GetOutputDcOffsetResponse.GetOutputDcOffsetResponseError
    def __init__(self, success: _Optional[_Union[GetOutputDcOffsetResponse.GetOutputDcOffsetResponseSuccess, _Mapping]] = ..., error: _Optional[_Union[GetOutputDcOffsetResponse.GetOutputDcOffsetResponseError, _Mapping]] = ...) -> None: ...

class MixInputsDcOffset(_message.Message):
    __slots__ = ("I", "Q")
    I_FIELD_NUMBER: _ClassVar[int]
    Q_FIELD_NUMBER: _ClassVar[int]
    I: float
    Q: float
    def __init__(self, I: _Optional[float] = ..., Q: _Optional[float] = ...) -> None: ...

class SingleInputDcOffset(_message.Message):
    __slots__ = ("offset",)
    OFFSET_FIELD_NUMBER: _ClassVar[int]
    offset: float
    def __init__(self, offset: _Optional[float] = ...) -> None: ...

class JobServiceGetConfigResponse(_message.Message):
    __slots__ = ("success", "error")
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    success: _qm_api_pb2.GetConfigSuccess
    error: _qm_api_pb2.GetConfigError
    def __init__(self, success: _Optional[_Union[_qm_api_pb2.GetConfigSuccess, _Mapping]] = ..., error: _Optional[_Union[_qm_api_pb2.GetConfigError, _Mapping]] = ...) -> None: ...

class JobServiceGetJobStatusResponse(_message.Message):
    __slots__ = ("success", "error")
    class JobServiceGetJobStatusResponseError(_message.Message):
        __slots__ = ("details",)
        DETAILS_FIELD_NUMBER: _ClassVar[int]
        details: str
        def __init__(self, details: _Optional[str] = ...) -> None: ...
    class JobServiceGetJobStatusResponseSuccess(_message.Message):
        __slots__ = ("status",)
        STATUS_FIELD_NUMBER: _ClassVar[int]
        status: _common_types_pb2.JobExecutionStatus
        def __init__(self, status: _Optional[_Union[_common_types_pb2.JobExecutionStatus, str]] = ...) -> None: ...
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    success: JobServiceGetJobStatusResponse.JobServiceGetJobStatusResponseSuccess
    error: JobServiceGetJobStatusResponse.JobServiceGetJobStatusResponseError
    def __init__(self, success: _Optional[_Union[JobServiceGetJobStatusResponse.JobServiceGetJobStatusResponseSuccess, _Mapping]] = ..., error: _Optional[_Union[JobServiceGetJobStatusResponse.JobServiceGetJobStatusResponseError, _Mapping]] = ...) -> None: ...

class JobServicePushToInputStreamRequest(_message.Message):
    __slots__ = ("job_id", "stream_name", "int_stream_data", "fixed_stream_data", "bool_stream_data")
    JOB_ID_FIELD_NUMBER: _ClassVar[int]
    STREAM_NAME_FIELD_NUMBER: _ClassVar[int]
    INT_STREAM_DATA_FIELD_NUMBER: _ClassVar[int]
    FIXED_STREAM_DATA_FIELD_NUMBER: _ClassVar[int]
    BOOL_STREAM_DATA_FIELD_NUMBER: _ClassVar[int]
    job_id: str
    stream_name: str
    int_stream_data: _job_manager_pb2.IntStreamData
    fixed_stream_data: _job_manager_pb2.FixedStreamData
    bool_stream_data: _job_manager_pb2.BoolStreamData
    def __init__(self, job_id: _Optional[str] = ..., stream_name: _Optional[str] = ..., int_stream_data: _Optional[_Union[_job_manager_pb2.IntStreamData, _Mapping]] = ..., fixed_stream_data: _Optional[_Union[_job_manager_pb2.FixedStreamData, _Mapping]] = ..., bool_stream_data: _Optional[_Union[_job_manager_pb2.BoolStreamData, _Mapping]] = ...) -> None: ...

class PushToInputStreamResponse(_message.Message):
    __slots__ = ("success", "error")
    class PushToInputStreamResponseSuccess(_message.Message):
        __slots__ = ()
        def __init__(self) -> None: ...
    class PushToInputStreamResponseError(_message.Message):
        __slots__ = ("details",)
        DETAILS_FIELD_NUMBER: _ClassVar[int]
        details: str
        def __init__(self, details: _Optional[str] = ...) -> None: ...
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    success: PushToInputStreamResponse.PushToInputStreamResponseSuccess
    error: PushToInputStreamResponse.PushToInputStreamResponseError
    def __init__(self, success: _Optional[_Union[PushToInputStreamResponse.PushToInputStreamResponseSuccess, _Mapping]] = ..., error: _Optional[_Union[PushToInputStreamResponse.PushToInputStreamResponseError, _Mapping]] = ...) -> None: ...

class GetJobResultSchemaRequest(_message.Message):
    __slots__ = ("job_id",)
    JOB_ID_FIELD_NUMBER: _ClassVar[int]
    job_id: str
    def __init__(self, job_id: _Optional[str] = ...) -> None: ...

class GetJobResultSchemaResponse(_message.Message):
    __slots__ = ("success", "error")
    class GetJobResultSchemaResponseSuccess(_message.Message):
        __slots__ = ("items",)
        class Item(_message.Message):
            __slots__ = ("name", "simple_dtype", "is_single", "expected_count", "shape")
            NAME_FIELD_NUMBER: _ClassVar[int]
            SIMPLE_DTYPE_FIELD_NUMBER: _ClassVar[int]
            IS_SINGLE_FIELD_NUMBER: _ClassVar[int]
            EXPECTED_COUNT_FIELD_NUMBER: _ClassVar[int]
            SHAPE_FIELD_NUMBER: _ClassVar[int]
            name: str
            simple_dtype: str
            is_single: bool
            expected_count: int
            shape: _containers.RepeatedScalarFieldContainer[int]
            def __init__(self, name: _Optional[str] = ..., simple_dtype: _Optional[str] = ..., is_single: bool = ..., expected_count: _Optional[int] = ..., shape: _Optional[_Iterable[int]] = ...) -> None: ...
        ITEMS_FIELD_NUMBER: _ClassVar[int]
        items: _containers.RepeatedCompositeFieldContainer[GetJobResultSchemaResponse.GetJobResultSchemaResponseSuccess.Item]
        def __init__(self, items: _Optional[_Iterable[_Union[GetJobResultSchemaResponse.GetJobResultSchemaResponseSuccess.Item, _Mapping]]] = ...) -> None: ...
    class GetJobResultSchemaResponseError(_message.Message):
        __slots__ = ("details",)
        DETAILS_FIELD_NUMBER: _ClassVar[int]
        details: str
        def __init__(self, details: _Optional[str] = ...) -> None: ...
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    success: GetJobResultSchemaResponse.GetJobResultSchemaResponseSuccess
    error: GetJobResultSchemaResponse.GetJobResultSchemaResponseError
    def __init__(self, success: _Optional[_Union[GetJobResultSchemaResponse.GetJobResultSchemaResponseSuccess, _Mapping]] = ..., error: _Optional[_Union[GetJobResultSchemaResponse.GetJobResultSchemaResponseError, _Mapping]] = ...) -> None: ...

class GetJobNamedResultHeaderRequest(_message.Message):
    __slots__ = ("job_id", "output_name", "flat_format")
    JOB_ID_FIELD_NUMBER: _ClassVar[int]
    OUTPUT_NAME_FIELD_NUMBER: _ClassVar[int]
    FLAT_FORMAT_FIELD_NUMBER: _ClassVar[int]
    job_id: str
    output_name: str
    flat_format: bool
    def __init__(self, job_id: _Optional[str] = ..., output_name: _Optional[str] = ..., flat_format: bool = ...) -> None: ...

class GetJobNamedResultsHeadersRequest(_message.Message):
    __slots__ = ("job_id", "outputs")
    class Output(_message.Message):
        __slots__ = ("output_name", "flat_format")
        OUTPUT_NAME_FIELD_NUMBER: _ClassVar[int]
        FLAT_FORMAT_FIELD_NUMBER: _ClassVar[int]
        output_name: str
        flat_format: bool
        def __init__(self, output_name: _Optional[str] = ..., flat_format: bool = ...) -> None: ...
    JOB_ID_FIELD_NUMBER: _ClassVar[int]
    OUTPUTS_FIELD_NUMBER: _ClassVar[int]
    job_id: str
    outputs: _containers.RepeatedCompositeFieldContainer[GetJobNamedResultsHeadersRequest.Output]
    def __init__(self, job_id: _Optional[str] = ..., outputs: _Optional[_Iterable[_Union[GetJobNamedResultsHeadersRequest.Output, _Mapping]]] = ...) -> None: ...

class GetJobNamedResultsHeadersResponse(_message.Message):
    __slots__ = ("success", "error")
    class GetJobNamedResultsHeadersResponseSuccess(_message.Message):
        __slots__ = ("headers",)
        class OutputHeader(_message.Message):
            __slots__ = ("count_so_far", "has_data_loss", "output_name", "simple_dtype", "shape")
            COUNT_SO_FAR_FIELD_NUMBER: _ClassVar[int]
            HAS_DATA_LOSS_FIELD_NUMBER: _ClassVar[int]
            OUTPUT_NAME_FIELD_NUMBER: _ClassVar[int]
            SIMPLE_DTYPE_FIELD_NUMBER: _ClassVar[int]
            SHAPE_FIELD_NUMBER: _ClassVar[int]
            count_so_far: int
            has_data_loss: bool
            output_name: str
            simple_dtype: str
            shape: _containers.RepeatedScalarFieldContainer[int]
            def __init__(self, count_so_far: _Optional[int] = ..., has_data_loss: bool = ..., output_name: _Optional[str] = ..., simple_dtype: _Optional[str] = ..., shape: _Optional[_Iterable[int]] = ...) -> None: ...
        HEADERS_FIELD_NUMBER: _ClassVar[int]
        headers: _containers.RepeatedCompositeFieldContainer[GetJobNamedResultsHeadersResponse.GetJobNamedResultsHeadersResponseSuccess.OutputHeader]
        def __init__(self, headers: _Optional[_Iterable[_Union[GetJobNamedResultsHeadersResponse.GetJobNamedResultsHeadersResponseSuccess.OutputHeader, _Mapping]]] = ...) -> None: ...
    class GetJobNamedResultsHeadersResponseError(_message.Message):
        __slots__ = ("details",)
        DETAILS_FIELD_NUMBER: _ClassVar[int]
        details: str
        def __init__(self, details: _Optional[str] = ...) -> None: ...
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    success: GetJobNamedResultsHeadersResponse.GetJobNamedResultsHeadersResponseSuccess
    error: GetJobNamedResultsHeadersResponse.GetJobNamedResultsHeadersResponseError
    def __init__(self, success: _Optional[_Union[GetJobNamedResultsHeadersResponse.GetJobNamedResultsHeadersResponseSuccess, _Mapping]] = ..., error: _Optional[_Union[GetJobNamedResultsHeadersResponse.GetJobNamedResultsHeadersResponseError, _Mapping]] = ...) -> None: ...

class GetJobNamedResultHeaderResponse(_message.Message):
    __slots__ = ("success", "error")
    class GetJobNamedResultHeaderResponseError(_message.Message):
        __slots__ = ("details",)
        DETAILS_FIELD_NUMBER: _ClassVar[int]
        details: str
        def __init__(self, details: _Optional[str] = ...) -> None: ...
    class GetJobNamedResultHeaderResponseSuccess(_message.Message):
        __slots__ = ("is_single", "count_so_far", "simple_dtype", "has_data_loss", "shape", "output_name")
        IS_SINGLE_FIELD_NUMBER: _ClassVar[int]
        COUNT_SO_FAR_FIELD_NUMBER: _ClassVar[int]
        SIMPLE_DTYPE_FIELD_NUMBER: _ClassVar[int]
        HAS_DATA_LOSS_FIELD_NUMBER: _ClassVar[int]
        SHAPE_FIELD_NUMBER: _ClassVar[int]
        OUTPUT_NAME_FIELD_NUMBER: _ClassVar[int]
        is_single: bool
        count_so_far: int
        simple_dtype: str
        has_data_loss: bool
        shape: _containers.RepeatedScalarFieldContainer[int]
        output_name: str
        def __init__(self, is_single: bool = ..., count_so_far: _Optional[int] = ..., simple_dtype: _Optional[str] = ..., has_data_loss: bool = ..., shape: _Optional[_Iterable[int]] = ..., output_name: _Optional[str] = ...) -> None: ...
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    success: GetJobNamedResultHeaderResponse.GetJobNamedResultHeaderResponseSuccess
    error: GetJobNamedResultHeaderResponse.GetJobNamedResultHeaderResponseError
    def __init__(self, success: _Optional[_Union[GetJobNamedResultHeaderResponse.GetJobNamedResultHeaderResponseSuccess, _Mapping]] = ..., error: _Optional[_Union[GetJobNamedResultHeaderResponse.GetJobNamedResultHeaderResponseError, _Mapping]] = ...) -> None: ...

class GetNamedResultsRequest(_message.Message):
    __slots__ = ("job_id", "outputs")
    class Output(_message.Message):
        __slots__ = ("output_name", "range")
        OUTPUT_NAME_FIELD_NUMBER: _ClassVar[int]
        RANGE_FIELD_NUMBER: _ClassVar[int]
        output_name: str
        range: _common_types_pb2.Range
        def __init__(self, output_name: _Optional[str] = ..., range: _Optional[_Union[_common_types_pb2.Range, _Mapping]] = ...) -> None: ...
    JOB_ID_FIELD_NUMBER: _ClassVar[int]
    OUTPUTS_FIELD_NUMBER: _ClassVar[int]
    job_id: str
    outputs: _containers.RepeatedCompositeFieldContainer[GetNamedResultsRequest.Output]
    def __init__(self, job_id: _Optional[str] = ..., outputs: _Optional[_Iterable[_Union[GetNamedResultsRequest.Output, _Mapping]]] = ...) -> None: ...

class GetNamedResultRequest(_message.Message):
    __slots__ = ("job_id", "output_name", "limit", "long_offset")
    JOB_ID_FIELD_NUMBER: _ClassVar[int]
    OUTPUT_NAME_FIELD_NUMBER: _ClassVar[int]
    LIMIT_FIELD_NUMBER: _ClassVar[int]
    LONG_OFFSET_FIELD_NUMBER: _ClassVar[int]
    job_id: str
    output_name: str
    limit: int
    long_offset: _wrappers_pb2.Int64Value
    def __init__(self, job_id: _Optional[str] = ..., output_name: _Optional[str] = ..., limit: _Optional[int] = ..., long_offset: _Optional[_Union[_wrappers_pb2.Int64Value, _Mapping]] = ...) -> None: ...

class GetNamedResultResponse(_message.Message):
    __slots__ = ("success", "error")
    class GetNamedResultResponseError(_message.Message):
        __slots__ = ("details",)
        DETAILS_FIELD_NUMBER: _ClassVar[int]
        details: str
        def __init__(self, details: _Optional[str] = ...) -> None: ...
    class GetNamedResultResponseSuccess(_message.Message):
        __slots__ = ("count_of_items", "data", "data_chunk", "data_summary", "output_name")
        class DataChunk(_message.Message):
            __slots__ = ("data",)
            DATA_FIELD_NUMBER: _ClassVar[int]
            data: bytes
            def __init__(self, data: _Optional[bytes] = ...) -> None: ...
        class DataSummary(_message.Message):
            __slots__ = ("count",)
            COUNT_FIELD_NUMBER: _ClassVar[int]
            count: int
            def __init__(self, count: _Optional[int] = ...) -> None: ...
        COUNT_OF_ITEMS_FIELD_NUMBER: _ClassVar[int]
        DATA_FIELD_NUMBER: _ClassVar[int]
        DATA_CHUNK_FIELD_NUMBER: _ClassVar[int]
        DATA_SUMMARY_FIELD_NUMBER: _ClassVar[int]
        OUTPUT_NAME_FIELD_NUMBER: _ClassVar[int]
        count_of_items: int
        data: bytes
        data_chunk: GetNamedResultResponse.GetNamedResultResponseSuccess.DataChunk
        data_summary: GetNamedResultResponse.GetNamedResultResponseSuccess.DataSummary
        output_name: str
        def __init__(self, count_of_items: _Optional[int] = ..., data: _Optional[bytes] = ..., data_chunk: _Optional[_Union[GetNamedResultResponse.GetNamedResultResponseSuccess.DataChunk, _Mapping]] = ..., data_summary: _Optional[_Union[GetNamedResultResponse.GetNamedResultResponseSuccess.DataSummary, _Mapping]] = ..., output_name: _Optional[str] = ...) -> None: ...
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    success: GetNamedResultResponse.GetNamedResultResponseSuccess
    error: GetNamedResultResponse.GetNamedResultResponseError
    def __init__(self, success: _Optional[_Union[GetNamedResultResponse.GetNamedResultResponseSuccess, _Mapping]] = ..., error: _Optional[_Union[GetNamedResultResponse.GetNamedResultResponseError, _Mapping]] = ...) -> None: ...

class GetJobErrorsRequest(_message.Message):
    __slots__ = ("job_id",)
    JOB_ID_FIELD_NUMBER: _ClassVar[int]
    job_id: str
    def __init__(self, job_id: _Optional[str] = ...) -> None: ...

class GetJobErrorsResponse(_message.Message):
    __slots__ = ("success", "error")
    class GetJobErrorsResponseError(_message.Message):
        __slots__ = ("details",)
        DETAILS_FIELD_NUMBER: _ClassVar[int]
        details: str
        def __init__(self, details: _Optional[str] = ...) -> None: ...
    class GetJobErrorsResponseSuccess(_message.Message):
        __slots__ = ("errors",)
        class ExecutionErrorSeverity(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
            __slots__ = ()
            ERROR: _ClassVar[GetJobErrorsResponse.GetJobErrorsResponseSuccess.ExecutionErrorSeverity]
            WARNING: _ClassVar[GetJobErrorsResponse.GetJobErrorsResponseSuccess.ExecutionErrorSeverity]
        ERROR: GetJobErrorsResponse.GetJobErrorsResponseSuccess.ExecutionErrorSeverity
        WARNING: GetJobErrorsResponse.GetJobErrorsResponseSuccess.ExecutionErrorSeverity
        class Error(_message.Message):
            __slots__ = ("errorCode", "errorSeverity", "message")
            ERRORCODE_FIELD_NUMBER: _ClassVar[int]
            ERRORSEVERITY_FIELD_NUMBER: _ClassVar[int]
            MESSAGE_FIELD_NUMBER: _ClassVar[int]
            errorCode: int
            errorSeverity: GetJobErrorsResponse.GetJobErrorsResponseSuccess.ExecutionErrorSeverity
            message: str
            def __init__(self, errorCode: _Optional[int] = ..., errorSeverity: _Optional[_Union[GetJobErrorsResponse.GetJobErrorsResponseSuccess.ExecutionErrorSeverity, str]] = ..., message: _Optional[str] = ...) -> None: ...
        ERRORS_FIELD_NUMBER: _ClassVar[int]
        errors: _containers.RepeatedCompositeFieldContainer[GetJobErrorsResponse.GetJobErrorsResponseSuccess.Error]
        def __init__(self, errors: _Optional[_Iterable[_Union[GetJobErrorsResponse.GetJobErrorsResponseSuccess.Error, _Mapping]]] = ...) -> None: ...
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    success: GetJobErrorsResponse.GetJobErrorsResponseSuccess
    error: GetJobErrorsResponse.GetJobErrorsResponseError
    def __init__(self, success: _Optional[_Union[GetJobErrorsResponse.GetJobErrorsResponseSuccess, _Mapping]] = ..., error: _Optional[_Union[GetJobErrorsResponse.GetJobErrorsResponseError, _Mapping]] = ...) -> None: ...

class CancelRequest(_message.Message):
    __slots__ = ("job_id",)
    JOB_ID_FIELD_NUMBER: _ClassVar[int]
    job_id: str
    def __init__(self, job_id: _Optional[str] = ...) -> None: ...

class CancelResponse(_message.Message):
    __slots__ = ("success", "error")
    class CancelJobError(_message.Message):
        __slots__ = ("details",)
        DETAILS_FIELD_NUMBER: _ClassVar[int]
        details: str
        def __init__(self, details: _Optional[str] = ...) -> None: ...
    class CancelJobSuccess(_message.Message):
        __slots__ = ()
        def __init__(self) -> None: ...
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    success: CancelResponse.CancelJobSuccess
    error: CancelResponse.CancelJobError
    def __init__(self, success: _Optional[_Union[CancelResponse.CancelJobSuccess, _Mapping]] = ..., error: _Optional[_Union[CancelResponse.CancelJobError, _Mapping]] = ...) -> None: ...

class GetJobResultStateRequest(_message.Message):
    __slots__ = ("job_id",)
    JOB_ID_FIELD_NUMBER: _ClassVar[int]
    job_id: str
    def __init__(self, job_id: _Optional[str] = ...) -> None: ...

class GetJobResultStateResponse(_message.Message):
    __slots__ = ("success", "error")
    class GetJobResultStateResponseError(_message.Message):
        __slots__ = ("details",)
        DETAILS_FIELD_NUMBER: _ClassVar[int]
        details: str
        def __init__(self, details: _Optional[str] = ...) -> None: ...
    class GetJobResultStateResponseSuccess(_message.Message):
        __slots__ = ("done", "closed", "has_dataloss")
        DONE_FIELD_NUMBER: _ClassVar[int]
        CLOSED_FIELD_NUMBER: _ClassVar[int]
        HAS_DATALOSS_FIELD_NUMBER: _ClassVar[int]
        done: bool
        closed: bool
        has_dataloss: bool
        def __init__(self, done: bool = ..., closed: bool = ..., has_dataloss: bool = ...) -> None: ...
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    success: GetJobResultStateResponse.GetJobResultStateResponseSuccess
    error: GetJobResultStateResponse.GetJobResultStateResponseError
    def __init__(self, success: _Optional[_Union[GetJobResultStateResponse.GetJobResultStateResponseSuccess, _Mapping]] = ..., error: _Optional[_Union[GetJobResultStateResponse.GetJobResultStateResponseError, _Mapping]] = ...) -> None: ...

class PullSamplesRequest(_message.Message):
    __slots__ = ("job_id", "include_analog", "include_digital")
    JOB_ID_FIELD_NUMBER: _ClassVar[int]
    INCLUDE_ANALOG_FIELD_NUMBER: _ClassVar[int]
    INCLUDE_DIGITAL_FIELD_NUMBER: _ClassVar[int]
    job_id: str
    include_analog: bool
    include_digital: bool
    def __init__(self, job_id: _Optional[str] = ..., include_analog: bool = ..., include_digital: bool = ...) -> None: ...

class PullSamplesResponse(_message.Message):
    __slots__ = ("success", "error")
    class PullSamplesResponseError(_message.Message):
        __slots__ = ("details",)
        DETAILS_FIELD_NUMBER: _ClassVar[int]
        details: str
        def __init__(self, details: _Optional[str] = ...) -> None: ...
    class PullSamplesResponseSuccess(_message.Message):
        __slots__ = ("controller", "fem_id", "port_id", "mode", "double_data", "boolean_data")
        class Mode(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
            __slots__ = ()
            ANALOG: _ClassVar[PullSamplesResponse.PullSamplesResponseSuccess.Mode]
            DIGITAL: _ClassVar[PullSamplesResponse.PullSamplesResponseSuccess.Mode]
        ANALOG: PullSamplesResponse.PullSamplesResponseSuccess.Mode
        DIGITAL: PullSamplesResponse.PullSamplesResponseSuccess.Mode
        class doubleData(_message.Message):
            __slots__ = ("mw", "lf")
            MW_FIELD_NUMBER: _ClassVar[int]
            LF_FIELD_NUMBER: _ClassVar[int]
            mw: PullSamplesResponse.PullSamplesResponseSuccess.mw
            lf: PullSamplesResponse.PullSamplesResponseSuccess.lf
            def __init__(self, mw: _Optional[_Union[PullSamplesResponse.PullSamplesResponseSuccess.mw, _Mapping]] = ..., lf: _Optional[_Union[PullSamplesResponse.PullSamplesResponseSuccess.lf, _Mapping]] = ...) -> None: ...
        class mw(_message.Message):
            __slots__ = ("ducId", "I", "Q")
            DUCID_FIELD_NUMBER: _ClassVar[int]
            I_FIELD_NUMBER: _ClassVar[int]
            Q_FIELD_NUMBER: _ClassVar[int]
            ducId: int
            I: _containers.RepeatedScalarFieldContainer[float]
            Q: _containers.RepeatedScalarFieldContainer[float]
            def __init__(self, ducId: _Optional[int] = ..., I: _Optional[_Iterable[float]] = ..., Q: _Optional[_Iterable[float]] = ...) -> None: ...
        class lf(_message.Message):
            __slots__ = ("items",)
            ITEMS_FIELD_NUMBER: _ClassVar[int]
            items: _containers.RepeatedScalarFieldContainer[float]
            def __init__(self, items: _Optional[_Iterable[float]] = ...) -> None: ...
        class booleanData(_message.Message):
            __slots__ = ("data",)
            class Data(_message.Message):
                __slots__ = ("item",)
                ITEM_FIELD_NUMBER: _ClassVar[int]
                item: _containers.RepeatedScalarFieldContainer[bool]
                def __init__(self, item: _Optional[_Iterable[bool]] = ...) -> None: ...
            DATA_FIELD_NUMBER: _ClassVar[int]
            data: _containers.RepeatedCompositeFieldContainer[PullSamplesResponse.PullSamplesResponseSuccess.booleanData.Data]
            def __init__(self, data: _Optional[_Iterable[_Union[PullSamplesResponse.PullSamplesResponseSuccess.booleanData.Data, _Mapping]]] = ...) -> None: ...
        CONTROLLER_FIELD_NUMBER: _ClassVar[int]
        FEM_ID_FIELD_NUMBER: _ClassVar[int]
        PORT_ID_FIELD_NUMBER: _ClassVar[int]
        MODE_FIELD_NUMBER: _ClassVar[int]
        DOUBLE_DATA_FIELD_NUMBER: _ClassVar[int]
        BOOLEAN_DATA_FIELD_NUMBER: _ClassVar[int]
        controller: str
        fem_id: int
        port_id: int
        mode: PullSamplesResponse.PullSamplesResponseSuccess.Mode
        double_data: PullSamplesResponse.PullSamplesResponseSuccess.doubleData
        boolean_data: PullSamplesResponse.PullSamplesResponseSuccess.booleanData
        def __init__(self, controller: _Optional[str] = ..., fem_id: _Optional[int] = ..., port_id: _Optional[int] = ..., mode: _Optional[_Union[PullSamplesResponse.PullSamplesResponseSuccess.Mode, str]] = ..., double_data: _Optional[_Union[PullSamplesResponse.PullSamplesResponseSuccess.doubleData, _Mapping]] = ..., boolean_data: _Optional[_Union[PullSamplesResponse.PullSamplesResponseSuccess.booleanData, _Mapping]] = ...) -> None: ...
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    success: PullSamplesResponse.PullSamplesResponseSuccess
    error: PullSamplesResponse.PullSamplesResponseError
    def __init__(self, success: _Optional[_Union[PullSamplesResponse.PullSamplesResponseSuccess, _Mapping]] = ..., error: _Optional[_Union[PullSamplesResponse.PullSamplesResponseError, _Mapping]] = ...) -> None: ...

class SetOscillatorFrequencyRequest(_message.Message):
    __slots__ = ("job_id", "qe", "new_frequency_hz", "update_component")
    class UpdateComponentSelection(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        both: _ClassVar[SetOscillatorFrequencyRequest.UpdateComponentSelection]
        upconverter: _ClassVar[SetOscillatorFrequencyRequest.UpdateComponentSelection]
        downconverter: _ClassVar[SetOscillatorFrequencyRequest.UpdateComponentSelection]
    both: SetOscillatorFrequencyRequest.UpdateComponentSelection
    upconverter: SetOscillatorFrequencyRequest.UpdateComponentSelection
    downconverter: SetOscillatorFrequencyRequest.UpdateComponentSelection
    JOB_ID_FIELD_NUMBER: _ClassVar[int]
    QE_FIELD_NUMBER: _ClassVar[int]
    NEW_FREQUENCY_HZ_FIELD_NUMBER: _ClassVar[int]
    UPDATE_COMPONENT_FIELD_NUMBER: _ClassVar[int]
    job_id: str
    qe: str
    new_frequency_hz: float
    update_component: SetOscillatorFrequencyRequest.UpdateComponentSelection
    def __init__(self, job_id: _Optional[str] = ..., qe: _Optional[str] = ..., new_frequency_hz: _Optional[float] = ..., update_component: _Optional[_Union[SetOscillatorFrequencyRequest.UpdateComponentSelection, str]] = ...) -> None: ...

class SetOscillatorFrequencyResponse(_message.Message):
    __slots__ = ("success", "error")
    class ChangeOscillatorFrequencyError(_message.Message):
        __slots__ = ("details",)
        DETAILS_FIELD_NUMBER: _ClassVar[int]
        details: str
        def __init__(self, details: _Optional[str] = ...) -> None: ...
    class ChangeOscillatorFrequencySuccess(_message.Message):
        __slots__ = ()
        def __init__(self) -> None: ...
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    success: SetOscillatorFrequencyResponse.ChangeOscillatorFrequencySuccess
    error: SetOscillatorFrequencyResponse.ChangeOscillatorFrequencyError
    def __init__(self, success: _Optional[_Union[SetOscillatorFrequencyResponse.ChangeOscillatorFrequencySuccess, _Mapping]] = ..., error: _Optional[_Union[SetOscillatorFrequencyResponse.ChangeOscillatorFrequencyError, _Mapping]] = ...) -> None: ...

class GetWaveformReportRequest(_message.Message):
    __slots__ = ("job_id",)
    JOB_ID_FIELD_NUMBER: _ClassVar[int]
    job_id: str
    def __init__(self, job_id: _Optional[str] = ...) -> None: ...

class GetWaveformReportResponse(_message.Message):
    __slots__ = ("success", "error")
    class GetWaveformReportSuccess(_message.Message):
        __slots__ = ("waveformReport",)
        WAVEFORMREPORT_FIELD_NUMBER: _ClassVar[int]
        waveformReport: _struct_pb2.Struct
        def __init__(self, waveformReport: _Optional[_Union[_struct_pb2.Struct, _Mapping]] = ...) -> None: ...
    class GetWaveformReportError(_message.Message):
        __slots__ = ("details",)
        DETAILS_FIELD_NUMBER: _ClassVar[int]
        details: str
        def __init__(self, details: _Optional[str] = ...) -> None: ...
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    success: GetWaveformReportResponse.GetWaveformReportSuccess
    error: GetWaveformReportResponse.GetWaveformReportError
    def __init__(self, success: _Optional[_Union[GetWaveformReportResponse.GetWaveformReportSuccess, _Mapping]] = ..., error: _Optional[_Union[GetWaveformReportResponse.GetWaveformReportError, _Mapping]] = ...) -> None: ...



from typing import Union, TypeVar, Sequence

from qm.grpc.qm.pb import inc_qua_pb2

# The public name for the QuaProgramAnyScalarExpression type.
MessageExpressionType = inc_qua_pb2.QuaProgram.AnyScalarExpression

MessageVarType = inc_qua_pb2.QuaProgram.VarRefExpression

T = TypeVar("T")
OneOrMore = Union[T, Sequence[T]]



import warnings
from typing import Union, cast

from qm.grpc.qm.pb import inc_qua_config_pb2
from qm.api.models.capabilities import QopCaps
from qm.utils.protobuf_utils import proto_repeated_to_list
from qm.program._dict_to_pb_converter.base_converter import BaseDictToPbConverter
from qm.type_hinting.config_types import (
    AnalogOutputFilterConfigType,
    AnalogOutputFilterConfigTypeQop33,
    AnalogOutputFilterConfigTypeQop35,
)


class AnalogOutputFiltersConverter(
    BaseDictToPbConverter[
        Union[AnalogOutputFilterConfigType, AnalogOutputFilterConfigTypeQop33, AnalogOutputFilterConfigTypeQop35],
        inc_qua_config_pb2.QuaConfig.AnalogOutputPortFilter,
    ]
):
    def convert(
        self,
        input_data: Union[
            AnalogOutputFilterConfigType, AnalogOutputFilterConfigTypeQop33, AnalogOutputFilterConfigTypeQop35
        ],
    ) -> inc_qua_config_pb2.QuaConfig.AnalogOutputPortFilter:
        return self._analog_output_port_filters_to_pb(input_data)

    @staticmethod
    def _validate_high_pass_param_in_qop35(data: AnalogOutputFilterConfigTypeQop35) -> None:
        if data.get("high_pass") is not None and data.get("exponential_dc_gain") is None:
            value = cast(AnalogOutputFilterConfigTypeQop33, data)["high_pass"]
            warnings.warn(
                f"Setting the `high_pass` to {value} is equivalent to setting the `exponential_dc_gain` field "
                f"to 0 and adding an exponential filter of (1, {value}). The `high_pass` field will be deprecated in QUA 2.0.",
                DeprecationWarning,
            )

    def _set_exponential_param(
        self,
        item: inc_qua_config_pb2.QuaConfig.AnalogOutputPortFilter,
        data_with_defaults: Union[AnalogOutputFilterConfigTypeQop33, AnalogOutputFilterConfigTypeQop35],
    ) -> None:
        if "exponential" in data_with_defaults:
            exponential = [
                inc_qua_config_pb2.QuaConfig.ExponentialParameters(amplitude=exp_params[0], time_constant=exp_params[1])
                for exp_params in data_with_defaults["exponential"]
            ]
            self._set_pb_attr_config_v2(item.iir, exponential, "exponential", "exponential_v2")

    def _set_high_pass_param(
        self,
        item: inc_qua_config_pb2.QuaConfig.AnalogOutputPortFilter,
        data_with_defaults: AnalogOutputFilterConfigTypeQop33,
    ) -> None:
        if "high_pass" in data_with_defaults:
            self._set_pb_attr_config_v2(
                item.iir,
                data_with_defaults["high_pass"],
                "high_pass",
                "high_pass_v2",
                allow_nones=True,
                create_container=inc_qua_config_pb2.QuaConfig.IirFilter.HighPassContainer,
            )

    def _analog_output_port_filters_qop33_to_pb(
        self,
        data: Union[AnalogOutputFilterConfigTypeQop33, AnalogOutputFilterConfigTypeQop35],
    ) -> inc_qua_config_pb2.QuaConfig.AnalogOutputPortFilter:
        default_schema: AnalogOutputFilterConfigTypeQop33 = {"feedforward": [], "exponential": [], "high_pass": None}
        data_with_defaults = self._apply_defaults(
            cast(AnalogOutputFilterConfigTypeQop33, data), default_schema=default_schema
        )

        item = inc_qua_config_pb2.QuaConfig.AnalogOutputPortFilter()
        self._set_pb_attr_config_v2(item, data_with_defaults.get("feedforward"), "feedforward", "feedforward_v2")
        self._set_exponential_param(item, data_with_defaults)
        self._set_high_pass_param(item, data_with_defaults)

        if self._capabilities.supports(QopCaps.exponential_dc_gain_filter):
            data_with_defaults_35 = self._apply_defaults(
                cast(AnalogOutputFilterConfigTypeQop35, data_with_defaults),
                default_schema={"exponential_dc_gain": None},
            )
            data_with_defaults_35 = cast(  # For mypy, we already did cast in the previous line
                AnalogOutputFilterConfigTypeQop35, data_with_defaults_35
            )
            self._validate_high_pass_param_in_qop35(data_with_defaults_35)
            self._set_exponential_dc_gain_param(item, data_with_defaults_35)
        else:
            self._validate_unsupported_params(
                data_with_defaults,
                unsupported_params=["exponential_dc_gain"],
                supported_params=["high_pass"],
                supported_from=QopCaps.exponential_dc_gain_filter.from_qop_version,
            )

        return item

    @staticmethod
    def _set_exponential_dc_gain_param(
        item: inc_qua_config_pb2.QuaConfig.AnalogOutputPortFilter,
        data_with_defaults: AnalogOutputFilterConfigTypeQop35,
    ) -> None:
        if "exponential_dc_gain" in data_with_defaults:
            exponential_dc_gain = data_with_defaults.get("exponential_dc_gain")
            item.iir.exponential_dc_gain.CopyFrom(
                inc_qua_config_pb2.QuaConfig.IirFilter.ExponentialDcGainContainer(value=exponential_dc_gain)
            )

    def _analog_output_port_filters_to_pb(
        self,
        data: Union[AnalogOutputFilterConfigType, AnalogOutputFilterConfigTypeQop33, AnalogOutputFilterConfigTypeQop35],
    ) -> inc_qua_config_pb2.QuaConfig.AnalogOutputPortFilter:

        if self._capabilities.supports(QopCaps.exponential_iir_filter):
            self._validate_unsupported_params(
                data,
                unsupported_params=["feedback"],
                supported_params=["high_pass", "exponential"],
                supported_until=QopCaps.exponential_iir_filter.from_qop_version,
            )
            return self._analog_output_port_filters_qop33_to_pb(cast(AnalogOutputFilterConfigTypeQop33, data))
        else:
            self._validate_unsupported_params(
                data,
                unsupported_params=["exponential", "high_pass"],
                supported_params=["feedback"],
                supported_from=QopCaps.exponential_iir_filter.from_qop_version,
            )

            data = cast(AnalogOutputFilterConfigType, data)
            return inc_qua_config_pb2.QuaConfig.AnalogOutputPortFilter(
                feedforward=data.get("feedforward", []), feedback=data.get("feedback", [])
            )

    def deconvert(
        self, output_data: inc_qua_config_pb2.QuaConfig.AnalogOutputPortFilter
    ) -> Union[AnalogOutputFilterConfigTypeQop35, AnalogOutputFilterConfigTypeQop33, AnalogOutputFilterConfigType]:
        if self._capabilities.supports(QopCaps.exponential_iir_filter):
            raw_exponential = (
                output_data.iir.exponential_v2.value
                if self._capabilities.supports(QopCaps.config_v2)
                else output_data.iir.exponential
            )
            exponential = [(exp_params.amplitude, exp_params.time_constant) for exp_params in raw_exponential]
            feedforward = (
                output_data.feedforward_v2.value
                if self._capabilities.supports(QopCaps.config_v2)
                else output_data.feedforward
            )

            ret33: AnalogOutputFilterConfigTypeQop33 = {
                "feedforward": proto_repeated_to_list(feedforward),
                "exponential": exponential,
            }

            if self._capabilities.supports(QopCaps.config_v2):
                # We handle both cases: the container being None (as likely returned by the Gateway),
                # and an initialized container with value=None.
                ret33["high_pass"] = (
                    output_data.iir.high_pass_v2.value if output_data.iir.HasField("high_pass_v2") else None
                )
            else:
                ret33["high_pass"] = output_data.iir.high_pass if output_data.iir.HasField("high_pass") else None

            if self._capabilities.supports(QopCaps.exponential_dc_gain_filter):
                exponential_dc_gain = (
                    output_data.iir.exponential_dc_gain.value
                    if output_data.iir.HasField("exponential_dc_gain")
                    and output_data.iir.exponential_dc_gain.HasField("value")
                    else None
                )
                ret35 = cast(AnalogOutputFilterConfigTypeQop35, {**ret33, "exponential_dc_gain": exponential_dc_gain})
                return ret35

            return ret33
        else:
            ret: AnalogOutputFilterConfigType = {
                "feedforward": proto_repeated_to_list(output_data.feedforward),
                "feedback": proto_repeated_to_list(output_data.feedback),
            }
            return ret



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: qm/pb/inc_qm_api.proto
# Protobuf Python Version: 5.28.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    28,
    1,
    '',
    'qm/pb/inc_qm_api.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from qm.grpc.qm.pb import inc_qua_config_pb2 as qm_dot_pb_dot_inc__qua__config__pb2
from qm.grpc.qm.pb import general_messages_pb2 as qm_dot_pb_dot_general__messages__pb2
from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x16qm/pb/inc_qm_api.proto\x12\x0eqm.grpc.qm_api\x1a\x1aqm/pb/inc_qua_config.proto\x1a\x1cqm/pb/general_messages.proto\x1a\x1cgoogle/protobuf/struct.proto\"\xd7\x16\n\x10HighQmApiRequest\x12,\n\x06\x63onfig\x18\x01 \x01(\x0b\x32\x18.qm.grpc.qm_api.QmConfigB\x02\x18\x01\x12\x37\n\x0cstrongConfig\x18\x02 \x01(\x0b\x32\x1d.qm.grpc.qua_config.QuaConfigB\x02\x18\x01\x12\x18\n\x10quantumMachineId\x18\x03 \x01(\t\x12G\n\rsetCorrection\x18\x33 \x01(\x0b\x32..qm.grpc.qm_api.HighQmApiRequest.SetCorrectionH\x00\x12\x45\n\x0csetFrequency\x18\x34 \x01(\x0b\x32-.qm.grpc.qm_api.HighQmApiRequest.SetFrequencyH\x00\x12G\n\x0bsetDcOffset\x18\x35 \x01(\x0b\x32,.qm.grpc.qm_api.HighQmApiRequest.SetDcOffsetB\x02\x18\x01H\x00\x12K\n\x0fsetDigitalRoute\x18\x36 \x01(\x0b\x32\x30.qm.grpc.qm_api.HighQmApiRequest.SetDigitalRouteH\x00\x12\x43\n\x0bsetIOValues\x18\x37 \x01(\x0b\x32,.qm.grpc.qm_api.HighQmApiRequest.SetIOValuesH\x00\x12M\n\x10setInputDcOffset\x18\x38 \x01(\x0b\x32\x31.qm.grpc.qm_api.HighQmApiRequest.SetInputDcOffsetH\x00\x12O\n\x11setOutputDcOffset\x18\x39 \x01(\x0b\x32\x32.qm.grpc.qm_api.HighQmApiRequest.SetOutputDcOffsetH\x00\x12S\n\x13setOutputFilterTaps\x18: \x01(\x0b\x32\x34.qm.grpc.qm_api.HighQmApiRequest.SetOutputFilterTapsH\x00\x12]\n\x18setDigitalInputThreshold\x18; \x01(\x0b\x32\x39.qm.grpc.qm_api.HighQmApiRequest.SetDigitalInputThresholdH\x00\x12[\n\x17setDigitalInputDeadtime\x18= \x01(\x0b\x32\x38.qm.grpc.qm_api.HighQmApiRequest.SetDigitalInputDeadtimeH\x00\x12[\n\x17setDigitalInputPolarity\x18? \x01(\x0b\x32\x38.qm.grpc.qm_api.HighQmApiRequest.SetDigitalInputPolarityH\x00\x1a\xab\x01\n\rSetCorrection\x12*\n\ncorrection\x18\x01 \x01(\x0b\x32\x16.qm.grpc.qm_api.Matrix\x12\x0c\n\x02qe\x18\n \x01(\tH\x00\x12H\n\x05mixer\x18\x0b \x01(\x0b\x32\x37.qm.grpc.qm_api.HighQmApiRequest.SetCorrectionMixerInfoH\x00\x42\x16\n\x14oneof_set_correction\x1a\xbe\x01\n\x16SetCorrectionMixerInfo\x12\r\n\x05mixer\x18\x01 \x01(\t\x12!\n\x15intermediateFrequency\x18\x02 \x01(\x04\x42\x02\x18\x01\x12\x17\n\x0bloFrequency\x18\x03 \x01(\x04\x42\x02\x18\x01\x12\x19\n\x11\x66requencyNegative\x18\x04 \x01(\x08\x12#\n\x1bintermediateFrequencyDouble\x18\x05 \x01(\x01\x12\x19\n\x11loFrequencyDouble\x18\x06 \x01(\x01\x1a\x42\n\x0cSetFrequency\x12\r\n\x05value\x18\x01 \x01(\x01\x12\x0c\n\x02qe\x18\n \x01(\tH\x00\x42\x15\n\x13oneof_set_frequency\x1aj\n\x0bSetDcOffset\x12\t\n\x01I\x18\x01 \x01(\x01\x12\t\n\x01Q\x18\x02 \x01(\x01\x12\x35\n\x02qe\x18\n \x01(\x0b\x32\'.qm.grpc.qm_api.HighQmApiRequest.QePortH\x00\x42\x0e\n\x0coneof_set_dc\x1ap\n\x11SetOutputDcOffset\x12\t\n\x01I\x18\x01 \x01(\x01\x12\t\n\x01Q\x18\x02 \x01(\x01\x12\x35\n\x02qe\x18\n \x01(\x0b\x32\'.qm.grpc.qm_api.HighQmApiRequest.QePortH\x00\x42\x0e\n\x0coneof_set_dc\x1a\xae\x01\n\x13SetOutputFilterTaps\x12G\n\x06\x66ilter\x18\x01 \x01(\x0b\x32\x37.qm.grpc.qm_api.HighQmApiRequest.AnalogOutputPortFilter\x12\x35\n\x02qe\x18\n \x01(\x0b\x32\'.qm.grpc.qm_api.HighQmApiRequest.QePortH\x00\x42\x17\n\x15oneof_set_filter_Taps\x1a?\n\x16\x41nalogOutputPortFilter\x12\x13\n\x0b\x66\x65\x65\x64\x66orward\x18\x01 \x03(\x01\x12\x10\n\x08\x66\x65\x65\x64\x62\x61\x63k\x18\x02 \x03(\x01\x1ai\n\x10SetInputDcOffset\x12\x0e\n\x06offset\x18\x01 \x01(\x01\x12\x35\n\x02qe\x18\n \x01(\x0b\x32\'.qm.grpc.qm_api.HighQmApiRequest.QePortH\x00\x42\x0e\n\x0coneof_set_dc\x1a\xd3\x01\n\x0fSetDigitalRoute\x12\r\n\x05value\x18\x01 \x01(\r\x12\x38\n\x05\x64\x65lay\x18\n \x01(\x0b\x32\'.qm.grpc.qm_api.HighQmApiRequest.QePortH\x00\x12\x39\n\x06\x62uffer\x18\x0b \x01(\x0b\x32\'.qm.grpc.qm_api.HighQmApiRequest.QePortH\x00\x12\r\n\x03tof\x18\x0c \x01(\tH\x00\x12\x12\n\x08smearing\x18\r \x01(\tH\x00\x42\x19\n\x17oneof_set_digital_route\x1a|\n\x0bSetIOValues\x12\r\n\x03\x61ll\x18\x01 \x01(\x08H\x00\x12G\n\x0eioValueSetData\x18\x33 \x03(\x0b\x32/.qm.grpc.qm_api.HighQmApiRequest.IOValueSetDataB\x0f\n\roneof_setTypeJ\x04\x08\x02\x10\x33\x1a\"\n\x06QePort\x12\n\n\x02qe\x18\x01 \x01(\t\x12\x0c\n\x04port\x18\x02 \x01(\t\x1aw\n\x0eIOValueSetData\x12\x11\n\tio_number\x18\x01 \x01(\x05\x12\x12\n\x08intValue\x18\x02 \x01(\x05H\x00\x12\x15\n\x0b\x64oubleValue\x18\x03 \x01(\x01H\x00\x12\x16\n\x0c\x62ooleanValue\x18\x04 \x01(\x08H\x00\x42\x0f\n\roneof_IO_type\x1a\x64\n\x18SetDigitalInputThreshold\x12\x35\n\x0b\x64igitalPort\x18\x01 \x01(\x0b\x32 .qm.grpc.qm_api.DigitalInputPort\x12\x11\n\tthreshold\x18\x02 \x01(\x01\x1a\x8c\x01\n\x17SetDigitalInputPolarity\x12\x35\n\x0b\x64igitalPort\x18\x01 \x01(\x0b\x32 .qm.grpc.qm_api.DigitalInputPort\x12:\n\x08polarity\x18\x02 \x01(\x0e\x32(.qm.grpc.qm_api.DigitalInputPortPolarity\x1a\x62\n\x17SetDigitalInputDeadtime\x12\x35\n\x0b\x64igitalPort\x18\x01 \x01(\x0b\x32 .qm.grpc.qm_api.DigitalInputPort\x12\x10\n\x08\x64\x65\x61\x64time\x18\x02 \x01(\rB\x0b\n\tapi_oneofJ\x04\x08\x04\x10\x33\"R\n\x10\x44igitalInputPort\x12\x16\n\x0e\x63ontrollerName\x18\x01 \x01(\t\x12\x12\n\nportNumber\x18\x02 \x01(\r\x12\x12\n\nfem_Number\x18\x03 \x01(\r\"W\n\x11HighQmApiResponse\x12\n\n\x02ok\x18\x01 \x01(\x08\x12\x36\n\x06\x65rrors\x18\x02 \x03(\x0b\x32&.qm.grpc.general_messages.ErrorMessage\"B\n\x08QmConfig\x12\x0f\n\x07version\x18\x01 \x01(\r\x12%\n\x04root\x18\x02 \x01(\x0b\x32\x17.google.protobuf.Struct\"<\n\x06Matrix\x12\x0b\n\x03v00\x18\x01 \x01(\x01\x12\x0b\n\x03v01\x18\x02 \x01(\x01\x12\x0b\n\x03v10\x18\x03 \x01(\x01\x12\x0b\n\x03v11\x18\x04 \x01(\x01\"3\n\rPortReference\x12\x12\n\ncontroller\x18\x01 \x01(\t\x12\x0e\n\x06number\x18\x02 \x01(\r*3\n\x18\x44igitalInputPortPolarity\x12\n\n\x06RISING\x10\x00\x12\x0b\n\x07\x46\x41LLING\x10\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'qm.pb.inc_qm_api_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_HIGHQMAPIREQUEST_SETCORRECTIONMIXERINFO'].fields_by_name['intermediateFrequency']._loaded_options = None
  _globals['_HIGHQMAPIREQUEST_SETCORRECTIONMIXERINFO'].fields_by_name['intermediateFrequency']._serialized_options = b'\030\001'
  _globals['_HIGHQMAPIREQUEST_SETCORRECTIONMIXERINFO'].fields_by_name['loFrequency']._loaded_options = None
  _globals['_HIGHQMAPIREQUEST_SETCORRECTIONMIXERINFO'].fields_by_name['loFrequency']._serialized_options = b'\030\001'
  _globals['_HIGHQMAPIREQUEST'].fields_by_name['config']._loaded_options = None
  _globals['_HIGHQMAPIREQUEST'].fields_by_name['config']._serialized_options = b'\030\001'
  _globals['_HIGHQMAPIREQUEST'].fields_by_name['strongConfig']._loaded_options = None
  _globals['_HIGHQMAPIREQUEST'].fields_by_name['strongConfig']._serialized_options = b'\030\001'
  _globals['_HIGHQMAPIREQUEST'].fields_by_name['setDcOffset']._loaded_options = None
  _globals['_HIGHQMAPIREQUEST'].fields_by_name['setDcOffset']._serialized_options = b'\030\001'
  _globals['_DIGITALINPUTPORTPOLARITY']._serialized_start=3392
  _globals['_DIGITALINPUTPORTPOLARITY']._serialized_end=3443
  _globals['_HIGHQMAPIREQUEST']._serialized_start=131
  _globals['_HIGHQMAPIREQUEST']._serialized_end=3034
  _globals['_HIGHQMAPIREQUEST_SETCORRECTION']._serialized_start=1170
  _globals['_HIGHQMAPIREQUEST_SETCORRECTION']._serialized_end=1341
  _globals['_HIGHQMAPIREQUEST_SETCORRECTIONMIXERINFO']._serialized_start=1344
  _globals['_HIGHQMAPIREQUEST_SETCORRECTIONMIXERINFO']._serialized_end=1534
  _globals['_HIGHQMAPIREQUEST_SETFREQUENCY']._serialized_start=1536
  _globals['_HIGHQMAPIREQUEST_SETFREQUENCY']._serialized_end=1602
  _globals['_HIGHQMAPIREQUEST_SETDCOFFSET']._serialized_start=1604
  _globals['_HIGHQMAPIREQUEST_SETDCOFFSET']._serialized_end=1710
  _globals['_HIGHQMAPIREQUEST_SETOUTPUTDCOFFSET']._serialized_start=1712
  _globals['_HIGHQMAPIREQUEST_SETOUTPUTDCOFFSET']._serialized_end=1824
  _globals['_HIGHQMAPIREQUEST_SETOUTPUTFILTERTAPS']._serialized_start=1827
  _globals['_HIGHQMAPIREQUEST_SETOUTPUTFILTERTAPS']._serialized_end=2001
  _globals['_HIGHQMAPIREQUEST_ANALOGOUTPUTPORTFILTER']._serialized_start=2003
  _globals['_HIGHQMAPIREQUEST_ANALOGOUTPUTPORTFILTER']._serialized_end=2066
  _globals['_HIGHQMAPIREQUEST_SETINPUTDCOFFSET']._serialized_start=2068
  _globals['_HIGHQMAPIREQUEST_SETINPUTDCOFFSET']._serialized_end=2173
  _globals['_HIGHQMAPIREQUEST_SETDIGITALROUTE']._serialized_start=2176
  _globals['_HIGHQMAPIREQUEST_SETDIGITALROUTE']._serialized_end=2387
  _globals['_HIGHQMAPIREQUEST_SETIOVALUES']._serialized_start=2389
  _globals['_HIGHQMAPIREQUEST_SETIOVALUES']._serialized_end=2513
  _globals['_HIGHQMAPIREQUEST_QEPORT']._serialized_start=2515
  _globals['_HIGHQMAPIREQUEST_QEPORT']._serialized_end=2549
  _globals['_HIGHQMAPIREQUEST_IOVALUESETDATA']._serialized_start=2551
  _globals['_HIGHQMAPIREQUEST_IOVALUESETDATA']._serialized_end=2670
  _globals['_HIGHQMAPIREQUEST_SETDIGITALINPUTTHRESHOLD']._serialized_start=2672
  _globals['_HIGHQMAPIREQUEST_SETDIGITALINPUTTHRESHOLD']._serialized_end=2772
  _globals['_HIGHQMAPIREQUEST_SETDIGITALINPUTPOLARITY']._serialized_start=2775
  _globals['_HIGHQMAPIREQUEST_SETDIGITALINPUTPOLARITY']._serialized_end=2915
  _globals['_HIGHQMAPIREQUEST_SETDIGITALINPUTDEADTIME']._serialized_start=2917
  _globals['_HIGHQMAPIREQUEST_SETDIGITALINPUTDEADTIME']._serialized_end=3015
  _globals['_DIGITALINPUTPORT']._serialized_start=3036
  _globals['_DIGITALINPUTPORT']._serialized_end=3118
  _globals['_HIGHQMAPIRESPONSE']._serialized_start=3120
  _globals['_HIGHQMAPIRESPONSE']._serialized_end=3207
  _globals['_QMCONFIG']._serialized_start=3209
  _globals['_QMCONFIG']._serialized_end=3275
  _globals['_MATRIX']._serialized_start=3277
  _globals['_MATRIX']._serialized_end=3337
  _globals['_PORTREFERENCE']._serialized_start=3339
  _globals['_PORTREFERENCE']._serialized_end=3390
# @@protoc_insertion_point(module_scope)



from qm.simulate._simulator_samples import SimulatorSamples, SimulatorControllerSamples

from ._streams_manager import StreamsManager
from ._single_stream_fetchers import (
    BaseSingleStreamFetcher,
    SingleStreamSingleResultFetcher,
    SingleStreamMultipleResultFetcher,
)

# Keeping these names for backwards compatibility
StreamingResultFetcher = StreamsManager
SingleStreamingResultFetcher = SingleStreamSingleResultFetcher
MultipleStreamingResultFetcher = SingleStreamMultipleResultFetcher

__all__ = [
    "StreamsManager",
    "StreamingResultFetcher",
    "BaseSingleStreamFetcher",
    "SingleStreamingResultFetcher",
    "MultipleStreamingResultFetcher",
    "SingleStreamSingleResultFetcher",
    "SingleStreamMultipleResultFetcher",
    "SimulatorSamples",
    "SimulatorControllerSamples",
]



import abc
import warnings
from typing import TYPE_CHECKING, Any, Type, Union, Generic, Literal, TypeVar, Optional, Sequence, overload

import numpy as np

from qm._loc import _get_loc
from qm.grpc.qm.pb import inc_qua_pb2
from qm.exceptions import QmQuaException
from qm.utils import deprecation_message
from qm.type_hinting.general import NumberT
from qm.serialization.expression_serializing_visitor import ExpressionSerializingVisitor

_ScalarExpressionType = inc_qua_pb2.QuaProgram.AnyScalarExpression


if TYPE_CHECKING:
    from qm.qua._qua_struct import _QuaStruct


def to_literal(
    value: Union[bool, int, float], dtype: inc_qua_pb2.QuaProgram.Type
) -> inc_qua_pb2.QuaProgram.LiteralExpression:
    return inc_qua_pb2.QuaProgram.LiteralExpression(value=str(value), type=dtype, loc=_get_loc())


def literal_int(value: int) -> _ScalarExpressionType:
    return inc_qua_pb2.QuaProgram.AnyScalarExpression(literal=to_literal(value, inc_qua_pb2.QuaProgram.Type.INT))


def literal_bool(value: bool) -> _ScalarExpressionType:
    return inc_qua_pb2.QuaProgram.AnyScalarExpression(literal=to_literal(value, inc_qua_pb2.QuaProgram.Type.BOOL))


def literal_real(value: float) -> _ScalarExpressionType:
    return inc_qua_pb2.QuaProgram.AnyScalarExpression(literal=to_literal(value, inc_qua_pb2.QuaProgram.Type.REAL))


def io(number: Literal[1, 2]) -> _ScalarExpressionType:
    return inc_qua_pb2.QuaProgram.AnyScalarExpression(
        variable=inc_qua_pb2.QuaProgram.VarRefExpression(ioNumber=number, loc=_get_loc())
    )


def io1() -> _ScalarExpressionType:
    return io(1)


def io2() -> _ScalarExpressionType:
    return io(2)


ScalarMessageType = TypeVar(
    "ScalarMessageType",
    inc_qua_pb2.QuaProgram.VarRefExpression,
    inc_qua_pb2.QuaProgram.LiteralExpression,
    inc_qua_pb2.QuaProgram.BinaryExpression,
    inc_qua_pb2.QuaProgram.ArrayCellRefExpression,
    inc_qua_pb2.QuaProgram.ArrayLengthExpression,
    inc_qua_pb2.QuaProgram.LibFunctionExpression,
    inc_qua_pb2.QuaProgram.FunctionExpression,
    inc_qua_pb2.QuaProgram.BroadcastExpression,
    inc_qua_pb2.QuaProgram.GlobalVarRefExpression,
)


S = TypeVar(
    "S",
    bound=Union[
        inc_qua_pb2.QuaProgram.ArrayVarRefExpression,
        inc_qua_pb2.QuaProgram.AnyScalarExpression,
        inc_qua_pb2.QuaProgram.StructVarRefExpression,
        inc_qua_pb2.QuaProgram.ExternalStreamRefExpression,
    ],
)


class QuaExpression(Generic[S], metaclass=abc.ABCMeta):
    def __init__(self, expression: S):
        self._expression = expression

    def unwrap(self) -> S:
        return self._expression

    @property
    def unwrapped(self) -> S:
        return self.unwrap()

    def __str__(self) -> str:
        return ExpressionSerializingVisitor(None).serialize(self._expression)

    def __bool__(self) -> bool:
        raise QmQuaException(
            "Attempted to use a Python logical operator on a QUA variable. If you are unsure why you got this message,"
            " please see https://docs.quantum-machines.co/latest/docs/Guides/qua_ref/#boolean-operations"
        )


class QuaNumericExpression(Generic[S, NumberT], QuaExpression[S], metaclass=abc.ABCMeta):
    def __init__(self, expression: S, t: Type[NumberT]):
        super().__init__(expression)
        self._type: Type[NumberT] = t

    @property
    def dtype(self) -> Type[NumberT]:
        return self._type

    @property
    def _qua_type(self) -> inc_qua_pb2.QuaProgram.Type:
        if issubclass(self.dtype, bool):
            return inc_qua_pb2.QuaProgram.Type.BOOL
        if issubclass(self.dtype, int):
            return inc_qua_pb2.QuaProgram.Type.INT
        if issubclass(self.dtype, float):
            return inc_qua_pb2.QuaProgram.Type.REAL
        raise NotImplementedError(f"Unsupported type - {self.dtype}")

    @property
    def _is_input_stream(self) -> bool:
        return isinstance(self, InputStreamOldInterface)

    def empty(self) -> bool:
        warnings.warn(
            deprecation_message(
                method="QuaExpression.empty()",
                deprecated_in="1.1.0",
                removed_in="1.2.0",
                details="This function is going to be removed, as it's not needed.",
            ),
            DeprecationWarning,
        )
        return self._expression is None

    def isFixed(self) -> bool:
        warnings.warn(
            deprecation_message(
                method="QuaVariable.isFixed",
                deprecated_in="1.1.0",
                removed_in="1.2.0",
                details="use: 'QuaVariable.is_fixed()' instead",
            ),
            DeprecationWarning,
        )
        return self.is_fixed()

    def isInt(self) -> bool:
        warnings.warn(
            deprecation_message(
                method="QuaVariable.isInt",
                deprecated_in="1.1.0",
                removed_in="1.2.0",
                details="use: 'QuaVariable.is_int()' instead",
            ),
            DeprecationWarning,
        )
        return self.is_int()

    def isBool(self) -> bool:
        warnings.warn(
            deprecation_message(
                method="QuaVariable.isBool",
                deprecated_in="1.1.0",
                removed_in="1.2.0",
                details="use: 'QuaVariable.is_bool()' instead",
            ),
            DeprecationWarning,
        )
        return self.is_bool()

    def is_fixed(self) -> bool:
        return issubclass(self._type, float)

    def is_int(self) -> bool:
        return self._type == int

    def is_bool(self) -> bool:
        return self._type == bool


class ScalarMessageInterface(Generic[ScalarMessageType], metaclass=abc.ABCMeta):
    @property
    @abc.abstractmethod
    def unwrapped_scalar(self) -> ScalarMessageType:
        pass


class QuaScalarExpression(
    QuaNumericExpression[inc_qua_pb2.QuaProgram.AnyScalarExpression, NumberT],
    ScalarMessageInterface[ScalarMessageType],
    metaclass=abc.ABCMeta,
):
    """
    A class representing a QUA scalar - could be a single value (like a variable), or the result of multiple operations
    (between several expressions) that result in a Scalar value.
    """

    def _get_binary_pb_expression(
        self,
        other: "ScalarOfAnyType",
        op: inc_qua_pb2.QuaProgram.BinaryExpression.BinaryOperator,
        self_is_first: bool = True,
    ) -> inc_qua_pb2.QuaProgram.AnyScalarExpression:
        other_as_exp = to_scalar_pb_expression(other)
        self_as_exp = self.unwrapped
        if self_is_first:
            left, right = self_as_exp, other_as_exp
        else:
            left, right = other_as_exp, self_as_exp

        exp = inc_qua_pb2.QuaProgram.AnyScalarExpression(
            binaryOperation=inc_qua_pb2.QuaProgram.BinaryExpression(loc=_get_loc(), left=left, right=right, op=op)
        )
        return exp

    def _binary(
        self,
        other: "Scalar[NumberT]",
        op: inc_qua_pb2.QuaProgram.BinaryExpression.BinaryOperator,
        self_is_first: bool = True,
    ) -> "QuaBinaryOperation[NumberT]":
        exp = self._get_binary_pb_expression(other, op, self_is_first)
        return QuaBinaryOperation(exp, self.dtype)

    def _boolean_binary(
        self,
        other: "Scalar[NumberT]",
        op: inc_qua_pb2.QuaProgram.BinaryExpression.BinaryOperator,
        self_is_first: bool = True,
    ) -> "QuaBinaryOperation[bool]":
        exp = self._get_binary_pb_expression(other, op, self_is_first)
        return QuaBinaryOperation(exp, bool)

    def _shift_operation_binary(
        self,
        other: "Scalar[int]",
        op: inc_qua_pb2.QuaProgram.BinaryExpression.BinaryOperator,
        self_is_first: bool = True,
    ) -> "QuaBinaryOperation[NumberT]":
        exp = self._get_binary_pb_expression(other, op, self_is_first)
        return QuaBinaryOperation(exp, self.dtype)

    def __add__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[NumberT]":
        return self._binary(other, inc_qua_pb2.QuaProgram.BinaryExpression.BinaryOperator.ADD)

    def __radd__(self, other: NumberT) -> "QuaBinaryOperation[NumberT]":
        return self._binary(other, inc_qua_pb2.QuaProgram.BinaryExpression.BinaryOperator.ADD, self_is_first=False)

    def __sub__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[NumberT]":
        return self._binary(other, inc_qua_pb2.QuaProgram.BinaryExpression.BinaryOperator.SUB)

    def __rsub__(self, other: NumberT) -> "QuaBinaryOperation[NumberT]":
        return self._binary(other, inc_qua_pb2.QuaProgram.BinaryExpression.BinaryOperator.SUB, self_is_first=False)

    def __neg__(self) -> "QuaBinaryOperation[NumberT]":
        return self.cast(0) - self

    def __gt__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[bool]":
        return self._boolean_binary(other, inc_qua_pb2.QuaProgram.BinaryExpression.BinaryOperator.GT)

    def __ge__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[bool]":
        return self._boolean_binary(other, inc_qua_pb2.QuaProgram.BinaryExpression.BinaryOperator.GET)

    def __lt__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[bool]":
        return self._boolean_binary(other, inc_qua_pb2.QuaProgram.BinaryExpression.BinaryOperator.LT)

    def __le__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[bool]":
        return self._boolean_binary(other, inc_qua_pb2.QuaProgram.BinaryExpression.BinaryOperator.LET)

    def __eq__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[bool]":  # type: ignore[override]
        return self._boolean_binary(other, inc_qua_pb2.QuaProgram.BinaryExpression.BinaryOperator.EQ)

    def __mul__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[NumberT]":
        return self._binary(other, inc_qua_pb2.QuaProgram.BinaryExpression.BinaryOperator.MULT)

    def __rmul__(self, other: NumberT) -> "QuaBinaryOperation[NumberT]":
        return self._binary(other, inc_qua_pb2.QuaProgram.BinaryExpression.BinaryOperator.MULT, self_is_first=False)

    def __truediv__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[NumberT]":
        return self._binary(other, inc_qua_pb2.QuaProgram.BinaryExpression.BinaryOperator.DIV)

    def __rtruediv__(self, other: NumberT) -> "QuaBinaryOperation[NumberT]":
        return self._binary(other, inc_qua_pb2.QuaProgram.BinaryExpression.BinaryOperator.DIV, self_is_first=False)

    def __lshift__(self, other: "Scalar[int]") -> "QuaBinaryOperation[NumberT]":
        return self._shift_operation_binary(other, inc_qua_pb2.QuaProgram.BinaryExpression.BinaryOperator.SHL)

    def __rlshift__(self, other: int) -> "QuaBinaryOperation[NumberT]":
        return self._shift_operation_binary(
            other, inc_qua_pb2.QuaProgram.BinaryExpression.BinaryOperator.SHL, self_is_first=False
        )

    def __rshift__(self, other: "Scalar[int]") -> "QuaBinaryOperation[NumberT]":
        return self._shift_operation_binary(other, inc_qua_pb2.QuaProgram.BinaryExpression.BinaryOperator.SHR)

    def __rrshift__(self, other: int) -> "QuaBinaryOperation[NumberT]":
        return self._shift_operation_binary(
            other, inc_qua_pb2.QuaProgram.BinaryExpression.BinaryOperator.SHR, self_is_first=False
        )

    def __and__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[NumberT]":
        return self._binary(other, inc_qua_pb2.QuaProgram.BinaryExpression.BinaryOperator.AND)

    def __rand__(self, other: NumberT) -> "QuaBinaryOperation[NumberT]":
        return self._binary(other, inc_qua_pb2.QuaProgram.BinaryExpression.BinaryOperator.AND, self_is_first=False)

    def __or__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[NumberT]":
        return self._binary(other, inc_qua_pb2.QuaProgram.BinaryExpression.BinaryOperator.OR)

    def __ror__(self, other: NumberT) -> "QuaBinaryOperation[NumberT]":
        return self._binary(other, inc_qua_pb2.QuaProgram.BinaryExpression.BinaryOperator.OR, self_is_first=False)

    def __xor__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[NumberT]":
        return self._binary(other, inc_qua_pb2.QuaProgram.BinaryExpression.BinaryOperator.XOR)

    def __rxor__(self, other: NumberT) -> "QuaBinaryOperation[NumberT]":
        return self._binary(other, inc_qua_pb2.QuaProgram.BinaryExpression.BinaryOperator.XOR, self_is_first=False)

    def __invert__(self) -> "QuaBinaryOperation[NumberT]":
        return self._binary(self.cast(True), inc_qua_pb2.QuaProgram.BinaryExpression.BinaryOperator.XOR)

    def cast(self, n: Union[bool, int, float]) -> NumberT:
        return self.dtype(n)

    @property
    def save_statement(self) -> inc_qua_pb2.QuaProgram.SaveStatement.Source:
        raise QmQuaException("saving is not allowed for this kind of qua expression")


class QuaArrayVariable(QuaNumericExpression[inc_qua_pb2.QuaProgram.ArrayVarRefExpression, NumberT]):
    def __init__(self, name: str, t: Type[NumberT], init_value: Sequence[Union[int, bool, float]], size: int):
        super(QuaArrayVariable, self).__init__(inc_qua_pb2.QuaProgram.ArrayVarRefExpression(name=name), t)
        self._size = size
        self._init_value = [to_literal(t(val), self._qua_type) for val in init_value]

    @property
    def declaration_statement(self) -> inc_qua_pb2.QuaProgram.VarDeclaration:
        return inc_qua_pb2.QuaProgram.VarDeclaration(
            name=self.unwrapped.name,
            value=self._init_value,
            type=self._qua_type,
            size=self._size,
            dim=1,
            isInputStream=self._is_input_stream,
        )

    def __getitem__(self, item: "Scalar[int]") -> "QuaArrayCell[NumberT]":
        idx_as_pb = to_scalar_pb_expression(item)
        arr = self.unwrapped
        loc = _get_loc()
        arr.loc = loc
        item_scalar_expression = inc_qua_pb2.QuaProgram.AnyScalarExpression(
            arrayCell=inc_qua_pb2.QuaProgram.ArrayCellRefExpression(arrayVar=arr, index=idx_as_pb, loc=loc)
        )
        return QuaArrayCell(item_scalar_expression, self.dtype)

    def length(self) -> "QuaArrayLength[int]":
        unwrapped_element = self.unwrapped
        array_exp = inc_qua_pb2.QuaProgram.ArrayLengthExpression(array=unwrapped_element)
        result = inc_qua_pb2.QuaProgram.AnyScalarExpression(arrayLength=array_exp)
        return QuaArrayLength(result, int)


class QuaStructReference(QuaExpression[inc_qua_pb2.QuaProgram.StructVarRefExpression]):
    def __init__(self, name: str):
        super().__init__(inc_qua_pb2.QuaProgram.StructVarRefExpression(name=name, loc=_get_loc()))
        self.name = name


NSize = TypeVar("NSize", bound=int)


class QuaStructArrayVariable(QuaArrayVariable[NumberT], Generic[NumberT, NSize]):
    def __init__(
        self,
        name: str,
        t: Type[NumberT],
        size: int,
        position: int,
        struct: QuaStructReference,
    ):
        super(QuaArrayVariable, self).__init__(
            inc_qua_pb2.QuaProgram.ArrayVarRefExpression(name=name, structVar=struct.unwrapped), t
        )
        self._size = size
        self._init_value = []
        self._position = position
        self._struct: QuaStructReference = struct

    @property
    def declaration_statement(self) -> inc_qua_pb2.QuaProgram.VarDeclaration:
        declaration_statement = super().declaration_statement
        struct_member = inc_qua_pb2.QuaProgram.VarDeclaration.StructMember(
            name=self._struct.name, position=self._position
        )
        declaration_statement.structMember.CopyFrom(struct_member)
        return declaration_statement


class AssignmentTargetInterface(metaclass=abc.ABCMeta):
    @property
    @abc.abstractmethod
    def assignment_statement(self) -> inc_qua_pb2.QuaProgram.AssignmentStatement.Target:
        pass


class NotAllowedOperation(QmQuaException):
    def __init__(self, expression_type: str) -> None:
        super(NotAllowedOperation, self).__init__(
            f"In-place operations are not supported for QUA {expression_type}. Please use `assign` instead."
        )


class NotAllowedOperationVariable(NotAllowedOperation):
    def __init__(self) -> None:
        super().__init__("variable")


class NotAllowedOperationArrayCell(NotAllowedOperation):
    def __init__(self) -> None:
        super().__init__("array cell")


class QuaVariable(AssignmentTargetInterface, QuaScalarExpression[NumberT, inc_qua_pb2.QuaProgram.VarRefExpression]):
    """A class representing a QUA scalar variable. Note that a `QuaVariable` is also a `QuaScalarExpression`."""

    def __init__(self, name: str, t: Type[NumberT], init_value: Optional[Union[int, bool, float]]):
        super(QuaScalarExpression, self).__init__(
            inc_qua_pb2.QuaProgram.AnyScalarExpression(variable=inc_qua_pb2.QuaProgram.VarRefExpression(name=name)), t
        )
        self._init_value = [to_literal(t(init_value), self._qua_type)] if init_value is not None else []

    @property
    def declaration_statement(self) -> inc_qua_pb2.QuaProgram.VarDeclaration:
        return inc_qua_pb2.QuaProgram.VarDeclaration(
            name=self.unwrapped.variable.name,
            value=self._init_value,
            type=self._qua_type,
            size=1,
            dim=0,
            isInputStream=self._is_input_stream,
        )

    @property
    def unwrapped_scalar(self) -> inc_qua_pb2.QuaProgram.VarRefExpression:
        return self.unwrapped.variable

    @property
    def save_statement(self) -> inc_qua_pb2.QuaProgram.SaveStatement.Source:
        return inc_qua_pb2.QuaProgram.SaveStatement.Source(variable=self.unwrapped_scalar)

    @property
    def assignment_statement(self) -> inc_qua_pb2.QuaProgram.AssignmentStatement.Target:
        return inc_qua_pb2.QuaProgram.AssignmentStatement.Target(variable=self.unwrapped_scalar)

    def __iadd__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[NumberT]":
        raise NotAllowedOperationVariable

    def __isub__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[NumberT]":
        raise NotAllowedOperationVariable

    def __imul__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[NumberT]":
        raise NotAllowedOperationVariable

    def __itruediv__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[NumberT]":
        raise NotAllowedOperationVariable

    def __ifloordiv__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[NumberT]":
        raise NotAllowedOperationVariable

    def __imod__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[NumberT]":
        raise NotAllowedOperationVariable

    def __ipow__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[NumberT]":
        raise NotAllowedOperationVariable

    def __ilshift__(self, other: "Scalar[int]") -> "QuaBinaryOperation[NumberT]":
        raise NotAllowedOperationVariable

    def __irshift__(self, other: "Scalar[int]") -> "QuaBinaryOperation[NumberT]":
        raise NotAllowedOperationVariable

    def __iand__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[NumberT]":
        raise NotAllowedOperationVariable

    def __ior__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[NumberT]":
        raise NotAllowedOperationVariable

    def __ixor__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[NumberT]":
        raise NotAllowedOperationVariable


class QuaLiteral(QuaScalarExpression[NumberT, inc_qua_pb2.QuaProgram.LiteralExpression]):
    @property
    def unwrapped_scalar(self) -> inc_qua_pb2.QuaProgram.LiteralExpression:
        return self.unwrapped.literal

    @property
    def save_statement(self) -> inc_qua_pb2.QuaProgram.SaveStatement.Source:
        return inc_qua_pb2.QuaProgram.SaveStatement.Source(literal=self.unwrapped_scalar)


class QuaArrayCell(
    AssignmentTargetInterface, QuaScalarExpression[NumberT, inc_qua_pb2.QuaProgram.ArrayCellRefExpression]
):
    """A class representing a QUA variable inside a QUA array cell."""

    @property
    def unwrapped_scalar(self) -> inc_qua_pb2.QuaProgram.ArrayCellRefExpression:
        return self.unwrapped.arrayCell

    @property
    def save_statement(self) -> inc_qua_pb2.QuaProgram.SaveStatement.Source:
        return inc_qua_pb2.QuaProgram.SaveStatement.Source(arrayCell=self.unwrapped_scalar)

    @property
    def assignment_statement(self) -> inc_qua_pb2.QuaProgram.AssignmentStatement.Target:
        return inc_qua_pb2.QuaProgram.AssignmentStatement.Target(arrayCell=self.unwrapped_scalar)

    def __iadd__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[NumberT]":
        raise NotAllowedOperationArrayCell

    def __isub__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[NumberT]":
        raise NotAllowedOperationArrayCell

    def __imul__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[NumberT]":
        raise NotAllowedOperationArrayCell

    def __itruediv__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[NumberT]":
        raise NotAllowedOperationArrayCell

    def __ifloordiv__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[NumberT]":
        raise NotAllowedOperationArrayCell

    def __imod__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[NumberT]":
        raise NotAllowedOperationArrayCell

    def __ipow__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[NumberT]":
        raise NotAllowedOperationArrayCell

    def __ilshift__(self, other: "Scalar[int]") -> "QuaBinaryOperation[NumberT]":
        raise NotAllowedOperationArrayCell

    def __irshift__(self, other: "Scalar[int]") -> "QuaBinaryOperation[NumberT]":
        raise NotAllowedOperationArrayCell

    def __iand__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[NumberT]":
        raise NotAllowedOperationArrayCell

    def __ior__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[NumberT]":
        raise NotAllowedOperationArrayCell

    def __ixor__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[NumberT]":
        raise NotAllowedOperationArrayCell


class QuaBinaryOperation(QuaScalarExpression[NumberT, inc_qua_pb2.QuaProgram.BinaryExpression]):
    @property
    def unwrapped_scalar(self) -> inc_qua_pb2.QuaProgram.BinaryExpression:
        return self.unwrapped.binaryOperation


class QuaArrayLength(QuaScalarExpression[NumberT, inc_qua_pb2.QuaProgram.ArrayLengthExpression]):
    @property
    def unwrapped_scalar(self) -> inc_qua_pb2.QuaProgram.ArrayLengthExpression:
        return self.unwrapped.arrayLength


class QuaLibFunctionOutput(QuaScalarExpression[NumberT, inc_qua_pb2.QuaProgram.LibFunctionExpression]):
    """A class representing the result of a QUA lib function."""

    @property
    def unwrapped_scalar(self) -> inc_qua_pb2.QuaProgram.LibFunctionExpression:
        return self.unwrapped.libFunction


class QuaFunctionOutput(QuaScalarExpression[NumberT, inc_qua_pb2.QuaProgram.FunctionExpression]):
    def __init__(self, function_expression: inc_qua_pb2.QuaProgram.FunctionExpression, t: Type[NumberT]):
        super(QuaScalarExpression, self).__init__(
            inc_qua_pb2.QuaProgram.AnyScalarExpression(function=function_expression), t
        )

    @property
    def unwrapped_scalar(self) -> inc_qua_pb2.QuaProgram.FunctionExpression:
        return self.unwrapped.function


class QuaBroadcast(QuaScalarExpression[NumberT, inc_qua_pb2.QuaProgram.BroadcastExpression]):
    """A class representing the result of a QUA broadcast expression."""

    def __init__(self, t: Type[NumberT], value: _ScalarExpressionType):
        super(QuaScalarExpression, self).__init__(
            inc_qua_pb2.QuaProgram.AnyScalarExpression(
                broadcast=inc_qua_pb2.QuaProgram.BroadcastExpression(value=value, loc=_get_loc())
            ),
            t,
        )

    @property
    def unwrapped_scalar(self) -> inc_qua_pb2.QuaProgram.BroadcastExpression:
        return self.unwrapped.broadcast


class StreamInterface(abc.ABC):
    @property
    @abc.abstractmethod
    def _direction(self) -> inc_qua_pb2.QuaProgram.Direction:
        pass


class OutputStreamInterface(StreamInterface, abc.ABC):
    @property
    def _direction(self) -> inc_qua_pb2.QuaProgram.Direction:
        return inc_qua_pb2.QuaProgram.Direction.OUTGOING


class InputStreamInterface(StreamInterface, abc.ABC):
    @property
    def _direction(self) -> inc_qua_pb2.QuaProgram.Direction:
        return inc_qua_pb2.QuaProgram.Direction.INCOMING


StructT = TypeVar("StructT", bound="_QuaStruct")


class QuaExternalStream(
    Generic[StructT], QuaExpression[inc_qua_pb2.QuaProgram.ExternalStreamRefExpression], StreamInterface
):
    def __init__(self, stream_id: int, struct_t: Type[StructT]):
        super(QuaExternalStream, self).__init__(
            inc_qua_pb2.QuaProgram.ExternalStreamRefExpression(stream_id=stream_id, loc=_get_loc())
        )
        self._stream_id = stream_id
        self._struct_t = struct_t

    @property
    def declaration_statement(self) -> inc_qua_pb2.QuaProgram.ExternalStreamDeclaration:
        return inc_qua_pb2.QuaProgram.ExternalStreamDeclaration(
            stream_id=self._stream_id,
            expectedTypes=self._struct_t.__underlying_declarations__,
            direction=self._direction,
        )


class QuaExternalIncomingStream(QuaExternalStream[StructT], InputStreamInterface):
    def receive(self, struct: StructT) -> None:
        # Alternative API to directly call `receive_from_opnic_stream`.
        # Importing `_receive_from_opnic_stream` here to avoid circular imports
        from qm.qua._dsl.streams.external_streams import _receive_from_opnic_stream

        _receive_from_opnic_stream(self, struct)


class QuaExternalOutgoingStream(QuaExternalStream[StructT], OutputStreamInterface):
    def send(self, struct: StructT) -> None:
        # Alternative API to directly call `send_to_opnic_stream`.
        # Importing `_send_to_opnic_stream` here to avoid circular imports
        from qm.qua._dsl.streams.external_streams import _send_to_opnic_stream

        _send_to_opnic_stream(self, struct)


class InputStreamOldInterface(InputStreamInterface, abc.ABC):
    @abc.abstractmethod
    def advance(self) -> inc_qua_pb2.QuaProgram.AnyStatement:
        # The statement itself, will be added to the program in the advance_input_stream() method
        pass


class QuaArrayInputStream(QuaArrayVariable[NumberT], InputStreamOldInterface):
    """A class representing the QUA vector that will be used as an input stream from the job to the QUA program."""

    def advance(self) -> inc_qua_pb2.QuaProgram.AnyStatement:
        return inc_qua_pb2.QuaProgram.AnyStatement(
            advanceInputStream=inc_qua_pb2.QuaProgram.AdvanceInputStreamStatement(
                loc=_get_loc(), streamArray=self.unwrapped
            )
        )


class QuaVariableInputStream(QuaVariable[NumberT], InputStreamOldInterface):
    """A class representing the QUA variable that will be used as an input stream from the job to the QUA program."""

    def advance(self) -> inc_qua_pb2.QuaProgram.AnyStatement:
        return inc_qua_pb2.QuaProgram.AnyStatement(
            advanceInputStream=inc_qua_pb2.QuaProgram.AdvanceInputStreamStatement(
                loc=_get_loc(), streamVariable=self.unwrapped.variable
            )
        )


class QuaIO(AssignmentTargetInterface):
    """A class representing the QUA IO type."""

    def __init__(self, number: Literal[1, 2]):
        self._number = number

    @property
    def assignment_statement(self) -> inc_qua_pb2.QuaProgram.AssignmentStatement.Target:
        return inc_qua_pb2.QuaProgram.AssignmentStatement.Target(
            variable=inc_qua_pb2.QuaProgram.VarRefExpression(ioNumber=self._number, loc=_get_loc())
        )


IO1 = QuaIO(1)
IO2 = QuaIO(2)


class _QuaGlobalVarOperation(
    QuaScalarExpression[NumberT, inc_qua_pb2.QuaProgram.GlobalVarRefExpression], metaclass=abc.ABCMeta
):
    def __init__(
        self, bits: Sequence[int], operation: inc_qua_pb2.QuaProgram.GlobalVarOperation, t: Type[NumberT]
    ) -> None:
        super(_QuaGlobalVarOperation, self).__init__(
            inc_qua_pb2.QuaProgram.AnyScalarExpression(
                globalVariable=inc_qua_pb2.QuaProgram.GlobalVarRefExpression(
                    loc=_get_loc(),
                    bits=list(bits),
                    operation=operation,
                ),
            ),
            t,
        )

    @property
    def unwrapped_scalar(self) -> inc_qua_pb2.QuaProgram.GlobalVarRefExpression:
        return self.unwrapped.globalVariable


class QuaGlobalVarRead(_QuaGlobalVarOperation[int]):
    def __init__(self, bits: Sequence[int], shift: bool) -> None:
        operation = (
            inc_qua_pb2.QuaProgram.GlobalVarOperation.read_shift
            if shift
            else inc_qua_pb2.QuaProgram.GlobalVarOperation.read
        )
        super(QuaGlobalVarRead, self).__init__(bits, operation, int)


class QuaGlobalVarXor(_QuaGlobalVarOperation[bool]):
    def __init__(self, bits: Sequence[int]) -> None:
        super(QuaGlobalVarXor, self).__init__(bits, inc_qua_pb2.QuaProgram.GlobalVarOperation.xor, bool)


def _fix_object_data_type(obj: Any) -> Any:
    if isinstance(obj, (np.floating, np.integer, np.bool_)):
        obj_item = obj.item()
        if isinstance(obj_item, np.longdouble):
            return float(obj_item)
        else:
            return obj_item
    else:
        return obj


def to_scalar_pb_expression(value: Union["ScalarOfAnyType", QuaIO]) -> _ScalarExpressionType:
    other = _fix_object_data_type(value)
    if isinstance(other, QuaScalarExpression):
        return other.unwrapped
    if isinstance(other, bool):  # Since bool is a subtype of int, it must be before it
        return literal_bool(other)
    if isinstance(other, int):
        return literal_int(other)
    if isinstance(other, float):
        return literal_real(other)
    if other == IO1:
        return io(1)
    if other == IO2:
        return io(2)
    raise QmQuaException(f"invalid expression: '{other}' is not a scalar expression")


@overload
def create_qua_scalar_expression(value: "QuaScalar[NumberT]") -> "QuaScalar[NumberT]":
    ...


@overload
def create_qua_scalar_expression(value: NumberT) -> QuaLiteral[NumberT]:
    ...


def create_qua_scalar_expression(value: "Scalar[NumberT]") -> "QuaScalar[NumberT]":
    if isinstance(value, QuaScalarExpression):
        return value
    if not isinstance(value, (int, float)):
        value = _fix_object_data_type(value)
    if isinstance(value, bool):  # Since bool is a subtype of int, it must be before it
        return QuaLiteral(literal_bool(value), bool)
    if isinstance(value, int):
        return QuaLiteral(literal_int(value), int)
    if isinstance(value, float):
        return QuaLiteral(literal_real(value), float)
    raise NotImplementedError


def validate_scalar_of_any_type(data_type: Any) -> None:
    if not isinstance(data_type, (QuaScalarExpression, bool, int, float)):
        raise QmQuaException(f"Data type must be a ScalarOfAnyType (QUA scalar value), got {type(data_type).__name__}.")


class fixed(float):
    pass


QuaScalar = Union[
    QuaVariable[NumberT],
    QuaLiteral[NumberT],
    QuaBinaryOperation[NumberT],
    QuaArrayCell[NumberT],
    QuaArrayLength[NumberT],
    QuaLibFunctionOutput[NumberT],
    QuaFunctionOutput[NumberT],
    QuaBroadcast[NumberT],
    _QuaGlobalVarOperation[NumberT],
]

Scalar = Union[QuaScalar[NumberT], NumberT]
"""A generic type representing the generic `NumberT` or the QUA equivalent of it."""

Vector = Union[Sequence[NumberT], QuaArrayVariable[NumberT]]
"""A generic type representing a generic array of `NumberT`, or the QUA equivalent of it."""

ScalarOfAnyType = Union[Scalar[bool], Scalar[int], Scalar[float]]
"""A type representing a scalar value in QUA, or the equivalent python type."""

VectorOfAnyType = Union[Vector[bool], Vector[int], Vector[float]]
"""A type representing a vector value in QUA, or the equivalent python type."""
